"""
Recallable per-channel lists of text items.

Commands:
    note-add <text>:
        Add a new note for this channel.
    note-edit <num> <text>:
        Update an existing note from this channel with new text.
    note-remove <num>:
        Delete an existing note from this channel by its position.
    note-show <num>:
        Recall a single note in this channel.
    note-list:
        Recall all notes for this channel.

.. note::
    This hook requires an active :class:`.DatabaseHook` to store data.
"""

import time

from peewee import CharField, IntegerField

import immp
from immp.hook.command import BadUsage, CommandParser, command
from immp.hook.database import BaseModel, DatabaseHook


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


class Note(BaseModel):
    """
    Representation of a single note.

    Attributes:
        timestamp (int):
            Creation time of the note.
        network (str):
            Network identifier for the channel's plug.
        channel (str):
            Channel identifier where the note was created.
        user (str):
            User identifier of the note's author.
        text (str):
            Note content.
    """

    timestamp = IntegerField(default=lambda: int(time.time()))
    network = CharField()
    channel = CharField(null=True)
    user = CharField(null=True)
    text = CharField()

    @classmethod
    def select_channel(cls, channel):
        return (cls.select().where(cls.network == channel.plug.network_id,
                                   cls.channel == channel.source)
                            .order_by(cls.timestamp))

    @classmethod
    def select_position(cls, channel, num):
        if num < 1:
            raise ValueError
        try:
            # ModelSelect.get() ignores the offset clause, use an index instead.
            return cls.select_channel(channel).limit(1).offset(num - 1)[0]
        except IndexError:
            raise Note.DoesNotExist from None

    @classmethod
    def select_position_multi(cls, channel, *nums):
        if any(num < 1 for num in nums):
            raise ValueError
        notes = list(cls.select_channel(channel))
        try:
            return [notes[num - 1] for num in nums]
        except IndexError:
            raise Note.DoesNotExist from None

    @property
    def ago(self):
        diff = int(time.time()) - self.timestamp
        for step, unit in ((60, "s"), (60, "m"), (24, "h")):
            if diff < step:
                return "{}{}".format(diff, unit)
            diff //= step
        return "{}d".format(diff)

    def __repr__(self):
        return "<{}: #{} {} @ {}: {}>".format(self.__class__.__name__, self.id, self.ago,
                                              repr(self.channel), repr(self.text))


class NotesHook(immp.Hook):
    """
    Hook for managing and recalling notes in channels.
    """

    schema = None

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.db = None

    async def start(self):
        self.db = self.host.resources[DatabaseHook].db
        self.db.create_tables([Note], safe=True)

    async def channel_migrate(self, old, new):
        count = (Note.update(network=new.plug.network_id, channel=new.source)
                     .where(Note.network == old.plug.network_id,
                            Note.channel == old.source).execute())
        return count > 0

    @command("note-add", parser=CommandParser.none)
    async def add(self, msg, text):
        """
        Add a new note for this channel.
        """
        Note.create(network=msg.channel.plug.network_id,
                    channel=msg.channel.source,
                    user=(msg.user.id or msg.user.username) if msg.user else None,
                    text=text.raw())
        count = Note.select_channel(msg.channel).count()
        await msg.channel.send(immp.Message(text="{} Added #{}".format(TICK, count)))

    @command("note-edit", parser=CommandParser.hybrid)
    async def edit(self, msg, num, text):
        """
        Update an existing note from this channel with new text.
        """
        try:
            note = Note.select_position(msg.channel, int(num))
        except ValueError:
            raise BadUsage from None
        except Note.DoesNotExist:
            text = "{} Does not exist".format(CROSS)
        else:
            note.text = text.raw()
            note.save()
            text = "{} Edited".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("note-remove")
    async def remove(self, msg, *nums):
        """
        Delete one or more notes from this channel by their positions.
        """
        if not nums:
            raise BadUsage
        try:
            nums = [int(num) for num in nums]
            notes = Note.select_position_multi(msg.channel, *nums)
        except ValueError:
            raise BadUsage from None
        except Note.DoesNotExist:
            text = "{} Does not exist".format(CROSS)
        else:
            count = Note.delete().where(Note.id.in_(tuple(note.id for note in notes))).execute()
            text = "{} Removed {} note{}".format(TICK, count, "" if count == 1 else "s")
        await msg.channel.send(immp.Message(text=text))

    @command("note-show")
    async def show(self, msg, num):
        """
        Recall a single note in this channel.
        """
        try:
            note = Note.select_position(msg.channel, int(num))
        except ValueError:
            raise BadUsage from None
        except Note.DoesNotExist:
            text = "{} Does not exist".format(CROSS)
        else:
            text = immp.RichText([immp.Segment("{}.".format(num), bold=True),
                                  immp.Segment("\t"),
                                  *immp.RichText.unraw(note.text, self.host),
                                  immp.Segment("\t"),
                                  immp.Segment(note.ago, italic=True)])
        await msg.channel.send(immp.Message(text=text))

    @command("note-list")
    async def list(self, msg, query=None):
        """
        Recall all notes for this channel, or search for text across all notes.
        """
        notes = Note.select_channel(msg.channel)
        if query:
            matches = notes.where(Note.text.contains(query))
            count = len(matches)
        else:
            count = len(notes)
        title = ("{}{} note{} in this channel{}"
                 .format(count, " matching" if query else "",
                         "" if count == 1 else "s", ":" if count else "."))
        text = immp.RichText([immp.Segment(title, bold=bool(notes))])
        for num, note in enumerate(notes, 1):
            if query and note not in matches:
                continue
            text.append(immp.Segment("\n"),
                        immp.Segment("{}.".format(num), bold=True),
                        immp.Segment("\t"),
                        *immp.RichText.unraw(note.text, self.host),
                        immp.Segment("\t"),
                        immp.Segment(note.ago, italic=True))
        await msg.channel.send(immp.Message(text=text))



"""
Self-serve Discord roles for users.

Config:
    roles ((str, int) dict):
        Mapping from user-facing role names to Discord role IDs.

Commands:
    role <name>:
        Claim a role of this name.
    unrole <name>:
        Drop the role of this name.
"""

import immp
from immp.hook.command import CommandParser, CommandScope, command
from immp.plug.discord import DiscordPlug


class _NoSuchRole(Exception):
    pass


class DiscordRoleHook(immp.Hook):
    """
    Hook to assign and unassign Discord roles to and from users.
    """

    schema = immp.Schema({"roles": {str: int}})

    def _common(self, msg, name):
        if name not in self.config["roles"]:
            raise _NoSuchRole
        client = msg.channel.plug._client
        channel = client.get_channel(int(msg.channel.source))
        member = channel.guild.get_member(int(msg.user.id))
        for role in channel.guild.roles:
            if role.id == self.config["roles"][name]:
                return role, member
        else:
            raise _NoSuchRole

    def _test(self, channel, user):
        return isinstance(channel.plug, DiscordPlug)

    @command("role", scope=CommandScope.shared, parser=CommandParser.none,
             test=_test, sync_aware=True)
    async def role(self, msg, name):
        try:
            role, member = self._common(msg, str(name))
        except _NoSuchRole:
            await msg.channel.send(immp.Message(text="No such role"))
            return
        else:
            await member.add_roles(role)
            await msg.channel.send(immp.Message(text="\N{WHITE HEAVY CHECK MARK} Added"))

    @command("unrole", scope=CommandScope.shared, parser=CommandParser.none,
             test=_test, sync_aware=True)
    async def unrole(self, msg, name):
        try:
            role, member = self._common(msg, str(name))
        except _NoSuchRole:
            await msg.channel.send(immp.Message(text="No such role"))
            return
        else:
            await member.remove_roles(role)
            await msg.channel.send(immp.Message(text="\N{WHITE HEAVY CHECK MARK} Removed"))



from copy import copy
from datetime import datetime, timezone
from enum import Enum
import re

from .error import PlugError
from .util import _no_escape, escape, pretty_str, unescape


@pretty_str
class User:
    """
    Generic class to represent senders of messages.

    Attributes:
        id (str):
            Plug-specific user identifier.
        plug (.Plug):
            Source plug instance, representing the domain this user comes from.
        username (str):
            User's chosen or allocated display name.
        real_name (str):
            User's preferred and/or family name.
        avatar (str):
            URL of the user's profile picture.
        link (str):
            Public profile URL, or identifier used for invites.
        suggested (bool):
            ``True`` if the display of this user is not important.  Useful for bot messages where
            displaying the author is superfluous -- plugs may choose to show the user only if
            required by the network to show something.
        raw:
            Optional plug-specific underlying user object.
    """

    __slots__ = ("_id", "plug", "username", "real_name", "avatar", "link", "suggested", "raw")

    def __init__(self, id_=None, plug=None, *, username=None, real_name=None, avatar=None,
                 link=None, suggested=False, raw=None):
        self.id = id_
        self.plug = plug
        self.username = username
        self.real_name = real_name
        self.avatar = avatar
        self.link = link
        self.suggested = suggested
        self.raw = raw

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = str(value) if value else None

    async def is_system(self):
        """
        Equivalent to :meth:`.Plug.user_is_system`.

        Returns:
            bool:
                ``True`` if the user relates to the plug itself.
        """
        return await self.plug.user_is_system(self)

    async def private_channel(self):
        """
        Equivalent to :meth:`.Plug.channel_for_user`.

        Returns:
            .Channel:
                Private channel for this user.
        """
        return await self.plug.channel_for_user(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.plug and not (other.plug and self.plug.name == other.plug.name):
            return False
        if self.id:
            return self.id == other.id
        else:
            return ((self.id, self.username, self.real_name, self.link) ==
                    (other.id, other.username, other.real_name, other.link))

    def __hash__(self):
        if self.id:
            return hash(self.id)
        else:
            return hash((self.username, self.real_name, self.link))

    def __repr__(self):
        return "<{}: {} {}>".format(self.__class__.__name__, repr(self.id),
                                    repr(self.real_name or self.username))


class Segment:
    """
    Substring of message text with consistent formatting.

    Calling :meth:`str` on an instance will return the segment text.  Similarly, the length of a
    segment is just the length of the contained text.

    Attributes:
        text (str):
            Plain segment text.
        bold (bool):
            Whether this segment should be formatted bold.
        italic (bool):
            Whether this segment should be emphasised.
        underline (bool):
            Whether this segment should be underlined.
        strike (bool):
            Whether this segment should be struck through.
        code (bool):
            Whether this segment should be monospaced.
        pre (bool):
            Whether this segment should be preformatted.
        link (str):
            Anchor URL if this segment represents a clickable link.
        mention (.User):
            Target user mentioned in this segment.
        plain (bool):
            ``True`` if the segment is not formatted.
    """

    __slots__ = ("text", "bold", "italic", "underline", "strike", "code", "pre", "link", "mention")

    _format_attrs = __slots__[1:]

    def __init__(self, text, *, bold=False, italic=False, underline=False, strike=False,
                 code=False, pre=False, link=None, mention=None):
        self.text = text
        self.bold = bold
        self.italic = italic
        self.underline = underline
        self.strike = strike
        self.code = code
        self.pre = pre
        self.link = link
        self.mention = mention

    def __len__(self):
        return len(self.text)

    @property
    def _tuple(self):
        return (self.text, self.bold, self.italic, self.underline, self.strike, self.code,
                self.pre, self.link, self.mention)

    @property
    def plain(self):
        return all(not getattr(self, attr) for attr in self._format_attrs)

    def same_format(self, other):
        """
        Test if this and another segment are equally formatted.  Works like ``==`` but ignores the
        segment text.

        Args:
            other (.Segment):
                Second segment for comparison.

        Returns:
            bool:
                ``True`` if the segments match.
        """
        return isinstance(other, self.__class__) and self._tuple[1:] == other._tuple[1:]

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._tuple == other._tuple

    def __hash__(self):
        return hash(self._tuple)

    def __str__(self):
        return self.text

    def __repr__(self):
        attrs = [" {}".format(attr) for attr in self._format_attrs if getattr(self, attr)]
        return "<{}: {}{}>".format(self.__class__.__name__, repr(self.text), "".join(attrs))


class RichText:
    """
    Common standard for formatted message text, akin to Hangouts' message segments.  This is a
    container designed to hold instances of :class:`.Segment`.

    Calling :meth:`str` on an instance will return the entire text without formatting.

    Instances may be indexed or sliced, which will apply to the segments.  The :attr:`slice.step`
    argument is overloaded to be boolean; if ``True``, slicing will apply to characters of the text
    content rather than the segments.  Length is always based on the content.

    Attributes:
        size (int):
            Number of segments in the text.
    """

    _tag_regex = re.compile(r"{}(.*?){}".format(_no_escape("<"), _no_escape(">")))
    _split_regex = re.compile(_no_escape(","))

    _bool_tags = Segment.__slots__[1:7]

    __slots__ = ("_segments",)

    def __init__(self, segments=None):
        self._segments = (list(segments) if segments else None) or []

    @property
    def size(self):
        return len(self._segments)

    def normalise(self):
        """
        Make a copy of this message with formatting applied from text boundaries.

        For example::

            Some<b> bold </>text.

        The bold boundaries would be moved to surround "bold" excluding the spaces::

            Some <b>bold</> text.

        Returns:
            .RichText:
                Normalised message text instance.
        """
        # Find leading or trailing whitespace in segments, and split them out.
        parts = []
        for segment in self._segments:
            if segment.plain:
                parts.append(segment)
                continue
            clone = copy(segment)
            match = re.match(r"(\s*)(.*?)(\s*)$", clone.text, re.DOTALL)
            before, clone.text, after = match.groups()
            if before:
                parts.append(Segment(before))
            if clone.text:
                parts.append(clone)
            if after:
                parts.append(Segment(after))
        # Look for segments like "<b>Left</>", " ", "<b>Right</>" (where two segments of equal
        # formatting are separated only by non-new-line whitespace) and combine them.
        merged = parts[:2]
        for right in parts[2:]:
            left, middle = merged[-2:]
            if left.same_format(right) and not middle.text.strip() and "\n" not in middle.text:
                left.text = "".join((left.text, middle.text, right.text))
                merged.pop()
            else:
                merged.append(right)
        # Look for adjacent segments with equal formatting and combine them.
        normal = [merged[0]]
        for segment in merged[1:]:
            prev = normal[-1]
            if segment.same_format(prev):
                prev.text += segment.text
            else:
                normal.append(segment)
        return RichText(normal)

    def clone(self):
        """
        Make a copy of this message text and all its segments.

        Returns:
            .RichText:
                Cloned message text instance.
        """
        return RichText([copy(segment) for segment in self._segments])

    def prepend(self, *segments):
        """
        Insert one or more segments at the beginning of the text.

        Args:
            segments (.Segment list):
                New segments to lead the message text.
        """
        self._segments = list(segments) + self._segments

    def append(self, *segments):
        """
        Insert one or more segments at the end of the text.

        Args:
            segments (.Segment list):
                New segments to tail the message text.
        """
        self._segments += segments

    def indent(self, chars="  "):
        """
        Prefix each line of text with another string.

        Args:
            chars (str):
                Prefix characters to prepend to each line.

        Returns:
            .RichText:
                Indented message text instance.
        """
        clone = RichText()
        if not self:
            return clone
        clone.append(Segment(chars))
        for segment in self:
            for i, line in enumerate(segment.text.split("\n")):
                if i > 0:
                    clone.append(Segment("\n{}".format(chars)))
                if line:
                    new = copy(segment)
                    new.text = line
                    clone.append(new)
        return clone

    def offset(self, pos):
        """
        Find the position within a segment corresponding to a point along the whole text.

        Args:
            pos (int):
                Position within the represented text.

        Returns:
            (int, int) tuple:
                Segment and offset within it.
        """
        if pos < 0 or pos > len(self):
            raise IndexError("Position {} out of bounds".format(pos))
        total = 0
        for i, segment in enumerate(self):
            total += len(segment)
            if total > pos:
                return i, len(segment) - (total - pos)
        else:
            return i + 1, 0

    def trim(self, length):
        """
        Reduce a long message text to a snippet with an ellipsis.

        Args:
            length (int):
                Maximum length of the message text.

        Returns:
            .RichText:
                Trimmed message text instance.
        """
        if len(self) <= length:
            return self
        clone = self[:length:True]
        clone[-1].text = clone[-1].text[:-2]
        clone.append(Segment("..."))
        return clone

    @classmethod
    def unraw(cls, text, host=None):
        """
        Inverse of :meth:`raw`, parse a string with formatting syntax into a rich instance.

        Args:
            text (str):
                Plain text with formatting syntax.
            host (.Host):
                Optional host instance, needed to resolve mention tags.

        Returns:
            .RichText:
                Parsed message text instance.
        """
        rich = cls()
        current = {}
        while True:
            match = cls._tag_regex.search(text)
            if not match:
                break
            start = match.start()
            end = match.end()
            last, text = text[:start], text[end:]
            if last or rich:
                rich.append(Segment(unescape(last, "<"), **current))
            current = {}
            tags = cls._split_regex.split(match.group(1))
            for tag in tags:
                # Bare link tags will use the segment text as the link, filled in at the end.
                for target in cls._bool_tags + ("link",):
                    if tag in (target, target[0]):
                        current[target] = True
                        break
                else:
                    if tag.startswith(("link=", "l=")):
                        current["link"] = unescape(tag.split("=", 1)[1], ",", ">")
                    elif tag.startswith(("mention=", "m=")) and host:
                        parts = re.split(_no_escape("/"), tag.split("=", 1)[1], 2)
                        plug = unescape(parts[0], "/", ",", ">")
                        user = unescape(parts[1], "/", ",", ">")
                        name = unescape(parts[2], ",", ">")
                        current["mention"] = User(id_=user, plug=host.plugs[plug],
                                                  real_name=name)
        if text:
            rich.append(Segment(unescape(text, "<"), **current))
        for segment in rich:
            if segment.link is True:
                segment.link = segment.text
        return rich

    def raw(self):
        """
        Serialise formatted text into a string representation, suitable for storage or transmission
        as plain text.

        Returns:
            str:
                Plain text with formatting syntax.
        """
        raw = ""
        last = "/"
        for segment in self:
            tags = []
            for tag in self._bool_tags:
                if getattr(segment, tag):
                    tags.append(tag[0])
            if segment.link:
                if segment.text == segment.link:
                    tags.append("l")
                else:
                    tags.append("l={}".format(escape(segment.link, ",", ">")))
            if segment.mention:
                tags.append("m={}/{}/{}"
                            .format(escape(segment.mention.plug.name, "/", ",", ">"),
                                    escape(segment.mention.id, "/", ",", ">"),
                                    escape(segment.mention.real_name or segment.mention.username,
                                           ",", ">")))
            current = ",".join(tags) or "/"
            if current != last:
                raw += "<{}>".format(current)
            last = current
            raw += escape(segment.text, "<")
        if last != "/":
            raw += "</>"
        return raw

    def __iter__(self):
        return iter(self._segments)

    def __len__(self):
        return sum(len(segment) for segment in self._segments)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None and not isinstance(key.step, bool):
                raise TypeError("RichText slice step must be boolean")
            elif not key.step:
                return RichText(self._segments[key.start:key.stop])
            start, end = key.start, key.stop
            if start is None and end is None:
                return self.clone()
            elif start is None:
                start = 0
            elif end is None:
                end = len(self)
            start_segment, start_offset = self.offset(start % len(self))
            end_segment, end_offset = self.offset(((end - 1) % len(self)) + 1)
            stop_segment = end_segment + 1 if end_offset else end_segment
            stop_offset = end_offset - start_offset if start_segment == end_segment else end_offset
            clone = RichText()
            for segment in self[start_segment:stop_segment]:
                clone.append(copy(segment))
            if start_offset:
                clone[0].text = clone[0].text[start_offset:]
            if end_offset:
                clone[-1].text = clone[-1].text[:stop_offset]
            return clone
        elif isinstance(key, int):
            return self._segments[key]
        else:
            raise TypeError("RichText indices must be integers or slices")

    def __add__(self, other):
        return RichText(self._segments + list(other))

    def __iadd__(self, other):
        self._segments += other
        return self

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and len(self) == len(other) and
                all(x == y for x, y in zip(self, other)))

    def __hash__(self):
        return hash(tuple(self._segments))

    def __str__(self):
        # Fallback implementation: just return the message text without formatting.
        return "".join(str(segment) for segment in self._segments)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, repr(self._segments))


class File:
    """
    Base file attachment object.

    Attributes:
        title (str):
            Name of file, if the plug supports names.
        type (.Type):
            Basic type of the file.
        source (str):
            Public URL to the original file location, if one is available.
    """

    __slots__ = ("title", "type", "source")

    class Type(Enum):
        """
        Possible file attachment types.

        Attributes:
            unknown:
                Default file type if not otherwise specified.
            image:
                Picture in a standard recognised format (e.g. PNG, JPEG).
        """
        unknown = 0
        image = 1

    def __init__(self, title=None, type_=Type.unknown, source=None):
        self.title = title
        self.type = type_
        self.source = source

    async def get_content(self, sess):
        """
        Stream the contents of the file, suitable for writing to a file or uploading elsewhere.

        The default implementation will try to fetch a file by the source field.  It may be
        overridden to add authentication or other metadata, but the method must remain callable
        with only a session passed to it.

        Args:
            sess (aiohttp.ClientSession):
                Existing HTTP session with which to make any requests.

        Returns:
            io.IOBase:
                Readable stream of the raw file.
        """
        if self.source:
            return await sess.get(self.source)
        else:
            raise PlugError("No accessible URL to the file")

    def __str__(self):
        if self.title and self.source:
            return "{} ({})".format(self.title, self.source)
        else:
            return self.title or self.source

    def __repr__(self):
        return "<{}: {} {}>".format(self.__class__.__name__, repr(self.title), self.type.name)


@pretty_str
class Location:
    """
    Attributes:
        latitude (float):
            North-South coordinate of the location in degrees.
        longitude (float):
            East-West coordinate of the location in degrees.
        coordintes (float):
            Read-only ``(latitude, longitude)`` pair.
        name (str):
            Name of the place represented by this location.
        address (str):
            Full street address of this place.
        google_map_url (str):
            URL to Google Maps centred on this place.
    """

    __slots__ = ("latitude", "longitude", "name", "address")

    def __init__(self, latitude=None, longitude=None, name=None, address=None):
        self.latitude = latitude
        self.longitude = longitude
        self.name = name
        self.address = address

    @property
    def coordinates(self):
        return (self.latitude, self.longitude)

    @property
    def google_map_url(self):
        return "https://www.google.com/maps/place/{},{}".format(self.latitude, self.longitude)

    def google_image_url(self, width, height=None):
        """
        Generate a static map image URL centred on this place.

        Args:
            width (int):
                Width of the image.
            height (int):
                Height of the image, matches the width (for a square image) if not specified.

        Returns:
            str:
                Corresponding image URL from the Google Maps API.
        """
        return ("https://maps.googleapis.com/maps/api/staticmap?center={0},{1}&"
                "markers=color:red%7C{0},{1}&size={2}x{3}"
                .format(self.latitude, self.longitude, width, height or width))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.coordinates == other.coordinates

    def __hash__(self):
        return hash(self.coordinates)

    def __str__(self):
        out = self.address or "{}, {}".format(*self.coordinates)
        if self.name:
            out = "{} ({})".format(self.name, out)
        return out

    def __repr__(self):
        return "<{}: {}, {}{}>".format(self.__class__.__name__, self.latitude, self.longitude,
                                       " {}".format(repr(self.name)) if self.name else "")


class _SentMessageSlots:

    # Multiple-inheritance compatibility: define attributes for both Message and Receipt here (i.e.
    # all fields needed for SentMessage).
    __slots__ = ("_text", "user", "action", "reply_to", "joined", "left", "title", "attachments",
                 "_id", "channel", "at", "revision", "edited", "deleted", "raw")


@pretty_str
class Message(_SentMessageSlots):
    """
    Base message content container, understood by all plugs.

    Attributes:
        text (.RichText):
            Representation of the message text content.
        user (.User):
            User profile that sent the message.
        action (bool):
            Whether this message should be presented as an action involving its user.
        reply_to (.Message):
            Parent message, which this message replies to.
        joined (.User list):
            Collection of users that just joined the channel.
        left (.User list):
            Collection of users that just parted the channel.
        title (str):
            New channel title, if this message represents a rename.
        attachments (.Attachment list):
            Additional data included in the message.
        raw:
            Optional plug-specific underlying message or event object.
    """

    __slots__ = ()

    def __init__(self, *, text=None, user=None, action=False, reply_to=None, joined=None,
                 left=None, title=None, attachments=None, raw=None):
        self.text = text
        self.user = user
        self.action = action
        self.reply_to = reply_to
        self.joined = joined or []
        self.left = left or []
        self.title = title
        self.attachments = attachments or []
        self.raw = raw

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        if value is None or isinstance(value, RichText):
            self._text = value
        elif isinstance(value, str):
            self._text = RichText([Segment(value)])
        else:
            raise ValueError("Message text must be RichText or plain str")

    def render(self, *, real_name=True, link_name=True, edit=False, delimiter="\n",
               quote_reply=False, strip_links=False, trim=None):
        """
        Add the sender's name (if present) to the start of the message text, suitable for sending
        as-is on plugs that need all the textual message content in the body.

        Args:
            real_name (bool):
                ``True`` (default) to display real names, or ``False`` to prefer usernames.  If
                only one kind of name is available, it will be used regardless of this setting.
            link_name (bool):
                ``True`` (default) to link the author's name to their profile, if a link exists.
            edit (bool):
                Whether this render should show an ``[edit]`` tag next to the author.
            delimiter (str):
                Characters added between the sender's name and the message text (a new line by
                default).
            quote_reply (bool):
                ``True`` to quote the parent message before the current one, prefixed with a
                box-drawing vertical line (not quoted by default).
            strip_links (bool):
                ``True`` to skip over links in the message text (links left intact by default).
            trim (int):
                Show an ellipsed snippet if the text exceeds this length, or ``None`` (default) for
                no trimming.

        Returns:
            .RichText:
                Rendered message body.
        """
        output = RichText()
        name = link = None
        action = self.action
        if self.user and not self.user.suggested:
            if real_name:
                name = self.user.real_name or self.user.username
            else:
                name = self.user.username or self.user.real_name
            if link_name:
                link = self.user.link
        if self.text:
            if isinstance(self.text, RichText):
                text = self.text
                if strip_links:
                    text = text.clone()
                    for segment in text:
                        segment.link = None
            else:
                text = RichText([Segment(self.text)])
            if trim:
                text = text.trim(trim)
            output.append(*text)
        elif self.attachments:
            action = True
            count = len(self.attachments)
            what = "{} files".format(count) if count > 1 else "this file"
            output.append(Segment("sent {}".format(what)))
        if name:
            if action:
                output.prepend(Segment(" "))
            else:
                output.prepend(Segment(delimiter))
                output.prepend(Segment(":", bold=True))
            if edit:
                output.prepend(Segment(" [edit]"))
            output.prepend(Segment(name, bold=True, link=link))
        elif edit:
            output.prepend(Segment("[edit] "))
        if action:
            for segment in output:
                segment.italic = True
        if quote_reply and self.reply_to:
            if isinstance(self.reply_to, Message):
                quoted = self.reply_to.render(real_name=real_name, link_name=link_name,
                                              delimiter=delimiter, strip_links=True, trim=32)
                output.prepend(*(quoted.indent("\N{BOX DRAWINGS LIGHT VERTICAL} ")), Segment("\n"))
        return output

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return ((self.text, self.user, self.action, self.reply_to) ==
                (other.text, other.user, other.action, other.reply_to))

    def __hash__(self):
        return hash((self.user, self.text, self.action, self.reply_to))

    def _repr_parts(self):
        return " ".join(filter(None, (repr(self.user) if self.user else None,
                                      repr(str(self.text)) if self.text else None,
                                      "join" if self.joined else None,
                                      "leave" if self.left else None,
                                      ("+{}".format(len(self.attachments))
                                       if self.attachments else None))))

    def __repr__(self):
        parts = self._repr_parts()
        return "<{}{}>".format(self.__class__.__name__, ": {}".format(parts) if parts else "")


@pretty_str
class Receipt(_SentMessageSlots):
    """
    Reference to a physical message received from a plug.  This provides metadata for identifying
    a source message, in addition to the actual content attributes.

    Attributes:
        id (str):
            Unique (to the plug) message identifier, which should persist across edits and deletes.
        channel (.Channel):
            Source channel of this message.
        at (datetime.datetime):
            Timestamp of the message according to the external server, defaults to the current time
            at creation if unset.
        revision (str):
            Key to uniquely identify updates to a previous message, defaults to :attr:`id`.  Need
            not be in the same format as the main identifier.
        edited (bool):
            Whether the message content has been changed.
        deleted (bool):
            Whether the message was deleted from its source.
    """

    __slots__ = ()

    def __init__(self, id_, channel, *, at=None, revision=None, edited=False, deleted=False):
        self.id = id_
        self.channel = channel
        self.at = at or datetime.now(timezone.utc)
        self.revision = str(revision or id)
        self.edited = edited
        self.deleted = deleted

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = str(value) if value else None

    async def delete(self):
        """
        Equivalent to :meth:`.Plug.delete`.
        """
        return await self.channel.plug.delete(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (self.id, self.revision, self.channel) == (other.id, other.revision, other.channel)

    def __hash__(self):
        return hash((self.id, self.revision, self.channel))

    def __repr__(self):
        return "<{}: {} @ {} {}{}{}>".format(self.__class__.__name__, self.id, self.at,
                                             repr(self.channel), " edited" if self.edited else "",
                                             " deleted" if self.deleted else "")


class SentMessage(Receipt, Message):
    """
    Combination of :class:`.Receipt` and :class:`.Message`.
    """

    __slots__ = ()

    def __init__(self, id_, channel, *, at=None, revision=None, edited=False, deleted=False,
                 text=None, user=None, action=False, reply_to=None, joined=None, left=None,
                 title=None, attachments=None, raw=None):
        Receipt.__init__(self, id_=id_, channel=channel, revision=revision, at=at, edited=edited,
                         deleted=deleted)
        Message.__init__(self, text=text, user=user, action=action, reply_to=reply_to,
                         joined=joined, left=left, title=title, attachments=attachments, raw=raw)

    def __repr__(self):
        parts = self._repr_parts()
        receipt = super().__repr__()
        return "{}{}{}".format(receipt[:-1], " {}".format(parts) if parts else "", receipt[-1])



"""
Fallback mention and word highlight support for plugs via private channels.

Mentions
~~~~~~~~

Config:
    plugs (str list):
        List of plug names to enable mention alerts for.
    usernames (bool):
        Whether to match network usernames (``True`` by default).
    real-names (bool):
        Whether to match user's display names (``False`` by default).
    ambiguous (bool):
        Whether to notify multiple potential users of an ambiguous mention (``False`` by default).

For networks that don't provide native user mentions, this plug can send users a private message
when mentioned by their username or real name.

A mention is matched from each ``@`` sign until whitespace is encountered.  For real names, spaces
and special characters are ignored, so that e.g. ``@fredbloggs`` will match *Fred Bloggs*.

Partial mentions are supported, failing any exact matches, by basic prefix search on real names.
For example, ``@fred`` will match *Frederick*, and ``@fredb`` will match *Fred Bloggs*.

Subscriptions
~~~~~~~~~~~~~

Config:
    plugs (str list):
        List of plug names to enable subscription alerts for.

Commands:
    sub-add <text>:
        Add a subscription to your trigger list.
    sub-remove <text>:
        Remove a subscription from your trigger list.
    sub-exclude <text>:
        Don't trigger a specific subscription in the current public channel.
    sub-list:
        Show all active subscriptions.

Allows users to opt in to private message notifications when chosen highlight words are used in a
group conversation.

.. note::
    This hook requires an active :class:`.DatabaseHook` to store data.
"""

from asyncio import wait
from collections import defaultdict
import logging
import re

from peewee import CharField, ForeignKeyField

import immp
from immp.hook.command import CommandScope, command
from immp.hook.database import BaseModel, DatabaseHook
from immp.hook.sync import SyncPlug


log = logging.getLogger(__name__)


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


class _Skip(Exception):
    # Message isn't applicable to the hook.
    pass


class SubTrigger(BaseModel):
    """
    Individual subscription trigger phrase for an individual user.

    Attributes:
        network (str):
            Network identifier that the user belongs to.
        user (str):
            User identifier as given by the plug.
        text (str):
            Subscription text that they wish to be notified on.
    """

    network = CharField()
    user = CharField()
    text = CharField()

    def __repr__(self):
        return "<{}: #{} {} ({} @ {})>".format(self.__class__.__name__, self.id, repr(self.text),
                                               repr(self.user), repr(self.network))


class SubExclude(BaseModel):
    """
    Exclusion for a trigger in a specific channel.

    Attributes:
        trigger (.SubTrigger):
            Containing trigger instance.
        network (str):
            Network identifier that the channel belongs to.
        user (str):
            Channel's own identifier.
    """

    trigger = ForeignKeyField(model=SubTrigger, backref="excludes")
    network = CharField()
    channel = CharField()

    def __repr__(self):
        return "<{}: #{} {} @ {} {}>".format(self.__class__.__name__, self.id, repr(self.network),
                                             repr(self.channel), repr(self.trigger))


class _AlertHookBase(immp.Hook):

    schema = immp.Schema({"groups": [str]})

    group = immp.Group.MergedProperty("groups")

    async def _get_members(self, msg):
        # Sync integration: avoid duplicate notifications inside and outside a synced channel.
        # Commands and excludes should apply to the sync, but notifications are based on the
        # network-native channel.
        if isinstance(msg.channel.plug, SyncPlug):
            # We're in the sync channel, so we've already handled this event in native channels.
            log.debug("Ignoring sync channel: %r", msg.channel)
            raise _Skip
        channel = msg.channel
        synced = SyncPlug.any_sync(self.host, msg.channel)
        if synced:
            # We're in the native channel of a sync, use this channel for reading config.
            log.debug("Translating sync channel: %r -> %r", msg.channel, synced)
            channel = synced
        members = [user for user in (await msg.channel.members()) or []
                   if self.group.has_plug(user.plug)]
        if not members:
            raise _Skip
        return channel, members


class MentionsHook(_AlertHookBase):
    """
    Hook to send mention alerts via private channels.
    """

    schema = immp.Schema({immp.Optional("usernames", True): bool,
                          immp.Optional("real-names", False): bool,
                          immp.Optional("ambiguous", False): bool}, _AlertHookBase.schema)

    @staticmethod
    def _clean(text):
        return re.sub(r"\W", "", text).lower() if text else None

    def match(self, mention, members):
        """
        Identify users relevant to a mention.

        Args:
            mention (str):
                Raw mention text, e.g. ``@fred``.
            members (.User list):
                List of members in the channel where the mention took place.

        Returns:
            .User set:
                All applicable members to be notified.
        """
        name = self._clean(mention)
        real_matches = set()
        real_partials = set()
        for member in members:
            if self.config["usernames"] and self._clean(member.username) == name:
                # Assume usernames are unique, only match the corresponding user.
                return {member}
            if self.config["real-names"]:
                real = self._clean(member.real_name)
                if real == name:
                    real_matches.add(member)
                if real.startswith(name):
                    real_partials.add(member)
        if real_matches:
            # Assume multiple identical real names is unlikely.
            # If it's the same person with two users, they both get mentioned.
            return real_matches
        elif len(real_partials) == 1 or self.config["ambiguous"]:
            # Return a single partial match if it exists.
            # Only allow multiple partials if enabled, else ignore the mention.
            return real_partials
        else:
            return set()

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not source.text or await sent.channel.is_private():
            return
        try:
            _, members = await self._get_members(sent)
        except _Skip:
            return
        mentioned = set()
        for mention in re.findall(r"@\S+", str(source.text)):
            matches = self.match(mention, members)
            if matches:
                log.debug("Mention %r applies: %r", mention, matches)
                mentioned.update(matches)
            else:
                log.debug("Mention %r doesn't apply", mention)
        for segment in source.text:
            if segment.mention and segment.mention in members:
                log.debug("Segment mention %r applies: %r", segment.text, segment.mention)
                mentioned.add(segment.mention)
        if not mentioned:
            return
        text = immp.RichText()
        if source.user:
            text.append(immp.Segment(source.user.real_name or source.user.username, bold=True),
                        immp.Segment(" mentioned you"))
        else:
            text.append(immp.Segment("You were mentioned"))
        title = await sent.channel.title()
        link = await sent.channel.link()
        if title:
            text.append(immp.Segment(" in "),
                        immp.Segment(title, italic=True))
        text.append(immp.Segment(":\n"))
        text += source.text
        if source.user and source.user.link:
            text.append(immp.Segment("\n"),
                        immp.Segment("Go to user", link=source.user.link))
        if link:
            text.append(immp.Segment("\n"),
                        immp.Segment("Go to channel", link=link))
        tasks = []
        for member in mentioned:
            if member == source.user:
                continue
            private = await sent.channel.plug.channel_for_user(member)
            if private:
                tasks.append(private.send(immp.Message(text=text)))
        if tasks:
            await wait(tasks)


class SubscriptionsHook(_AlertHookBase):
    """
    Hook to send trigger word alerts via private channels.
    """

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.db = None

    @classmethod
    def _clean(cls, text):
        return re.sub(r"[^\w ]", "", text).lower()

    async def start(self):
        self.db = self.host.resources[DatabaseHook].db
        self.db.create_tables([SubTrigger, SubExclude], safe=True)

    def _test(self, channel, user):
        return self.group.has_plug(channel.plug)

    @command("sub-add", scope=CommandScope.private, test=_test)
    async def add(self, msg, *words):
        """
        Add a subscription to your trigger list.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        _, created = SubTrigger.get_or_create(network=msg.channel.plug.network_id,
                                              user=msg.user.id, text=text)
        resp = "{} {}".format(TICK, "Subscribed" if created else "Already subscribed")
        await msg.channel.send(immp.Message(text=resp))

    @command("sub-remove", scope=CommandScope.private, test=_test)
    async def remove(self, msg, *words):
        """
        Remove a subscription from your trigger list.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        count = SubTrigger.delete().where(SubTrigger.network == msg.channel.plug.network_id,
                                          SubTrigger.user == msg.user.id,
                                          SubTrigger.text == text).execute()
        resp = "{} {}".format(TICK, "Unsubscribed" if count else "Not subscribed")
        await msg.channel.send(immp.Message(text=resp))

    @command("sub-list", scope=CommandScope.private, test=_test)
    async def list(self, msg):
        """
        Show all active subscriptions.
        """
        subs = SubTrigger.select().where(SubTrigger.network == msg.user.plug.network_id,
                                         SubTrigger.user == msg.user.id).order_by(SubTrigger.text)
        if subs:
            text = immp.RichText([immp.Segment("Your subscriptions:", bold=True)])
            for sub in subs:
                text.append(immp.Segment("\n- {}".format(sub.text)))
        else:
            text = "No active subscriptions."
        await msg.channel.send(immp.Message(text=text))

    @command("sub-exclude", scope=CommandScope.shared, test=_test)
    async def exclude(self, msg, *words):
        """
        Don't trigger a specific subscription in the current channel.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        try:
            trigger = SubTrigger.get(network=msg.user.plug.network_id,
                                     user=msg.user.id, text=text)
        except SubTrigger.DoesNotExist:
            resp = "{} Not subscribed".format(CROSS)
        else:
            exclude, created = SubExclude.get_or_create(trigger=trigger,
                                                        network=msg.channel.plug.network_id,
                                                        channel=msg.channel.source)
            if not created:
                exclude.delete_instance()
            resp = "{} {}".format(TICK, "Excluded" if created else "No longer excluded")
        await msg.channel.send(immp.Message(text=resp))

    @staticmethod
    def match(text, channel, present):
        """
        Identify users subscribed to text snippets in a message.

        Args:
            text (str):
                Cleaned message text.
            channel (.Channel):
                Channel where the subscriptions were triggered.
            present (((str, str), .User) dict):
                Mapping from network/user IDs to members of the source channel.

        Returns:
            (.User, str set) dict:
                Mapping from applicable users to their filtered triggers.
        """
        subs = set()
        for sub in SubTrigger.select():
            key = (sub.network, sub.user)
            if key in present and sub.text in text:
                subs.add(sub)
        triggered = defaultdict(set)
        excludes = set(SubExclude.select().where(SubExclude.trigger << subs,
                                                 SubExclude.network == channel.plug.network_id,
                                                 SubExclude.channel == channel.source))
        for trigger in subs - set(exclude.trigger for exclude in excludes):
            triggered[present[(trigger.network, trigger.user)]].add(trigger.text)
        return triggered

    async def channel_migrate(self, old, new):
        count = (SubExclude.update(network=new.plug.network_id, channel=new.source)
                           .where(SubExclude.network == old.plug.network_id,
                                  SubExclude.channel == old.source).execute())
        return count > 0

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not source.text or await sent.channel.is_private():
            return
        try:
            lookup, members = await self._get_members(sent)
        except _Skip:
            return
        present = {(member.plug.network_id, str(member.id)): member for member in members}
        triggered = self.match(self._clean(str(source.text)), lookup, present)
        if not triggered:
            return
        tasks = []
        for member, triggers in triggered.items():
            if member == source.user:
                continue
            private = await member.private_channel()
            if not private:
                continue
            text = immp.RichText()
            mentioned = immp.Segment(", ".join(sorted(triggers)), italic=True)
            if source.user:
                text.append(immp.Segment(source.user.real_name or source.user.username, bold=True),
                            immp.Segment(" mentioned "), mentioned)
            else:
                text.append(mentioned, immp.Segment(" mentioned"))
            title = await sent.channel.title()
            link = await sent.channel.link()
            if title:
                text.append(immp.Segment(" in "),
                            immp.Segment(title, italic=True))
            text.append(immp.Segment(":\n"))
            text += source.text
            if source.user and source.user.link:
                text.append(immp.Segment("\n"),
                            immp.Segment("Go to user", link=source.user.link))
            if link:
                text.append(immp.Segment("\n"),
                            immp.Segment("Go to channel", link=link))
            tasks.append(private.send(immp.Message(text=text)))
        if tasks:
            await wait(tasks)



from asyncio import CancelledError, Task, ensure_future, gather, wait
from collections import defaultdict
from itertools import chain
import logging
from operator import attrgetter

from .channel import Channel, Group
from .error import ConfigError
from .hook import Hook, ResourceHook
from .plug import Plug
from .stream import PlugStream
from .util import OpenState, pretty_str


log = logging.getLogger(__name__)


class HostGetter:
    """
    Filter property used to return a subset of objects by type.
    """

    def __init__(self, cls):
        self._cls = cls

    def __get__(self, instance, owner):
        if instance:
            return {name: obj for name, obj in instance._objects.items()
                    if isinstance(obj, self._cls)}
        else:
            return self


@pretty_str
class Host:
    """
    Main class responsible for starting, stopping, and interacting with plugs.

    To run as the main coroutine for an application, use :meth:`run` in conjunction with
    :meth:`.AbstractEventLoop.run_until_complete`.  To stop a running host in an async context,
    await :meth:`quit`.

    For finer control, you can call :meth:`open`, :meth:`process` and :meth:`close` explicitly.

    Attributes:
        plugs ((str, .Plug) dict):
            Collection of all registered plug instances, keyed by name.
        channels ((str, .Channel) dict):
            Collection of all registered channels, keyed by name.
        groups ((str, .Group) dict):
            Collection of all registered group, keyed by name.
        hooks ((str, .Hook) dict):
            Collection of all registered hooks, keyed by name.
        plain_hooks ((str, .Hook) dict):
            As above, but excluding resources.
        resources ((class, .ResourceHook) dict):
            Collection of all registered resource hooks, keyed by class.
        priority ((str, int) dict):
            Mapping from hook names to their configured ordering.  Unordered (lowest priority)
            hooks are not present.
        running (bool):
            Whether messages from plugs are being processed by the host.
    """

    __slots__ = ("_objects", "_resources", "_priority", "_loaded", "_stream", "_process")

    def __init__(self):
        self._objects = {}
        self._resources = {}
        self._priority = {}
        self._loaded = False
        self._stream = self._process = None

    plugs = HostGetter(Plug)
    channels = HostGetter(Channel)
    groups = HostGetter(Group)
    hooks = HostGetter(Hook)
    resources = property(lambda self: dict(self._resources))
    priority = property(lambda self: dict(self._priority))

    @property
    def plain_hooks(self):
        return {name: hook for name, hook in self.hooks.items()
                if not isinstance(hook, ResourceHook)}

    def ordered_hooks(self):
        """
        Sort all registered hooks by priority.

        Returns:
            (.Hook set) list:
                Hooks grouped by their relative order -- one set for each ascending priority,
                followed by unordered hooks at the end.
        """
        prioritised = defaultdict(set)
        rest = set()
        for hook in chain(self._resources.values(), self.plain_hooks.values()):
            if hook.state != OpenState.active:
                continue
            try:
                prioritised[self._priority[hook.name]].add(hook)
            except KeyError:
                rest.add(hook)
        ordered = [hooks for _, hooks in sorted(prioritised.items())]
        if rest:
            ordered.append(rest)
        return ordered

    def __contains__(self, key):
        return key in self._objects

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._objects[key]
        elif isinstance(key, type) and issubclass(key, ResourceHook):
            return self._resources[key]
        else:
            raise TypeError(key)

    @property
    def running(self):
        # This is the "public" status that external code can query.
        return self._stream is not None

    def add_plug(self, plug):
        """
        Register a plug to the host.

        If the host is already running, use :meth:`join_plug` instead, which will start the plug
        and begin processing messages straight away.

        Args:
            plug (.Plug):
                Existing plug instance to add.

        Returns:
            str:
                Name used to reference this plug.
        """
        if not isinstance(plug, Plug):
            raise TypeError(plug)
        elif plug.name in self._objects:
            raise ConfigError("Plug name '{}' already registered".format(plug.name))
        log.info("Adding plug: %r (%s)", plug.name, plug.__class__.__name__)
        self._objects[plug.name] = plug
        if self._loaded:
            plug.on_load()
        if self._stream:
            self._stream.add(plug)
        return plug.name

    def remove_plug(self, name):
        """
        Unregister an existing plug.

        .. warning::
            This will not notify any hooks with a reference to this plug, nor will it attempt to
            remove it from their state.

        Args:
            name (str):
                Name of a previously registered plug instance to remove.

        Returns:
            .Plug:
                Removed plug instance.
        """
        if name not in self.plugs:
            raise RuntimeError("Plug '{}' not registered to host".format(name))
        log.info("Removing plug: %s", name)
        plug = self._objects.pop(name)
        for label, channel in list(self.channels.items()):
            if channel.plug == plug:
                self.remove_channel(label)
        if self._stream:
            self._stream.remove(plug)
        return plug

    def add_channel(self, name, channel):
        """
        Register a channel to the host.  The channel's plug must be registered first.

        Args:
            name (str):
                Unique identifier for this channel to be referenced by plugs and hooks.
            channel (.Channel):
                Existing channel instance to add.
        """
        if not isinstance(channel, Channel):
            raise TypeError(channel)
        elif name in self._objects:
            raise ConfigError("Channel name '{}' already registered".format(name))
        log.info("Adding channel: %r (%s/%s)", name, channel.plug.name, channel.source)
        self._objects[name] = channel

    def remove_channel(self, name):
        """
        Unregister an existing channel.

        Args:
            name (str):
                Name of a previously registered channel instance to remove.
        """
        if name not in self.channels:
            raise RuntimeError("Channel '{}' not registered to host".format(name))
        log.info("Removing channel: %s", name)
        return self._objects.pop(name)

    def add_group(self, group):
        """
        Register a group to the host.

        Args:
            group (.Group):
                Existing group instance to add.

        Returns:
            str:
                Name used to reference this group.
        """
        if not isinstance(group, Group):
            raise TypeError(group)
        elif group.name in self._objects:
            raise ConfigError("Group name '{}' already registered".format(group.name))
        log.info("Adding group: %s", group.name)
        self._objects[group.name] = group
        return group.name

    def remove_group(self, name):
        """
        Unregister an existing group.

        .. warning::
            This will not notify any hooks with a reference to this group, nor will it attempt to
            remove it from their state.

        Args:
            name (str):
                Name of a previously registered group instance to remove.

        Returns:
            .Group:
                Removed group instance.
        """
        if name not in self.groups:
            raise RuntimeError("Group '{}' not registered to host".format(name))
        log.info("Removing group: %s", name)
        return self._objects.pop(name)

    def add_hook(self, hook, priority=None):
        """
        Register a hook to the host.

        Args:
            hook (.Hook):
                Existing hook instance to add.
            priority (int):
                Optional ordering constraint, applied to send and receive events.  Hooks registered
                with a priority will be processed in ascending priority order, followed by those
                without prioritisation.  Where multiple hooks share a priority value, events may be
                processed in parallel (e.g. on-receive is dispatched to all hooks simultaneously).

        Returns:
            str:
                Name used to reference this hook.
        """
        if not isinstance(hook, Hook):
            raise TypeError(hook)
        elif hook.name in self._objects:
            raise ConfigError("Hook name '{}' already registered".format(hook.name))
        elif hook.__class__ in self._resources:
            raise ConfigError("Resource class '{}' already registered"
                              .format(hook.__class__.__name__))
        elif priority is not None and priority in self._priority:
            raise ConfigError("Priority {} already registered".format(priority))
        log.info("Adding hook: %r (%s)", hook.name, hook.__class__.__name__)
        self._objects[hook.name] = hook
        if isinstance(hook, ResourceHook):
            log.info("Adding resource: %r (%s)", hook.name, hook.__class__.__name__)
            self._resources[hook.__class__] = hook
        if priority is not None:
            self._priority[hook.name] = priority
        if self._loaded:
            hook.on_load()
        return hook.name

    def prioritise_hook(self, name, priority):
        """
        Re-prioritise an existing hook.

        Args:
            name (str):
                Name of a previously registered hook instance to prioritise.
            priority (int):
                Optional ordering constraint -- see :meth:`add_hook`.
        """
        if name not in self.hooks:
            raise RuntimeError("Hook '{}' not registered to host".format(name))
        if priority is None:
            self._priority.pop(name, None)
        elif isinstance(priority, int) and priority >= 1:
            self._priority[name] = priority
        else:
            raise ValueError("Hook priority must be a positive integer")

    def remove_hook(self, name):
        """
        Unregister an existing hook.

        .. warning::
            This will not notify any dependent hooks with a reference to this hook, nor will it
            attempt to remove it from their state.

        Args:
            name (str):
                Name of a previously registered hook instance to remove.

        Returns:
            .Hook:
                Removed hook instance.
        """
        if name not in self.hooks:
            raise RuntimeError("Hook '{}' not registered to host".format(name))
        self._priority.pop(name, None)
        log.info("Removing hook: %s", name)
        hook = self._objects.pop(name)
        if isinstance(hook, ResourceHook):
            log.info("Removing resource: %r (%s)", name, hook.__class__.__name__)
            del self._resources[hook.__class__]
        return hook

    def loaded(self):
        """
        Trigger the on-load event for all plugs and hooks.
        """
        for hook in self._resources.values():
            hook.on_load()
        for plug in self.plugs.values():
            plug.on_load()
        for hook in self.plain_hooks.values():
            hook.on_load()
        self._loaded = True

    async def _try_state(self, state, objs, timeout=None):
        if not objs:
            return
        action = "open" if state == OpenState.active else "close"
        getter = attrgetter(action)
        tasks = {Task(getter(obj)()): obj for obj in objs}
        done, pending = await wait(tasks.keys(), timeout=timeout)
        for task in done:
            exc = task.exception()
            if exc:
                obj = tasks[task]
                log.error("Failed to %s %r", action, obj.name, exc_info=exc)
        for task in pending:
            obj = tasks[task]
            log.warning("Failed to %s %r after %s seconds", action, obj.name, timeout)

    async def open(self):
        """
        Connect all open plugs and start all hooks.
        """
        if not self._loaded:
            raise RuntimeError("On-load event must be sent before opening")
        log.debug("Opening resources")
        await self._try_state(OpenState.active, self._resources.values(), 30)
        log.debug("Opening plugs")
        await self._try_state(OpenState.active, self.plugs.values(), 30)
        log.debug("Opening remaining hooks")
        await self._try_state(OpenState.active, self.plain_hooks.values(), 30)

    async def close(self):
        """
        Disconnect all open plugs and stop all hooks.
        """
        log.debug("Closing non-resource hooks")
        await self._try_state(OpenState.inactive, self.plain_hooks.values(), 30)
        log.debug("Closing plugs")
        await self._try_state(OpenState.inactive, self.plugs.values(), 30)
        log.debug("Closing resources")
        await self._try_state(OpenState.inactive, self._resources.values(), 30)

    async def _safe_receive(self, hook, sent, source, primary):
        try:
            await hook.on_receive(sent, source, primary)
        except Exception:
            log.exception("Hook %r failed on-receive event", hook.name)

    async def _callback(self, sent, source, primary):
        ordered = self.ordered_hooks()
        for hooks in ordered:
            for hook in hooks:
                try:
                    result = await hook.before_receive(sent, source, primary)
                except Exception:
                    log.exception("Hook %r failed before-receive event", hook.name)
                    continue
                if result:
                    sent = result
                else:
                    # Message has been suppressed by a hook.
                    return
        for hooks in ordered:
            await gather(*(self._safe_receive(hook, sent, source, primary) for hook in hooks))

    async def channel_migrate(self, old, new):
        """
        Issue a migration call to all hooks.

        Args:
            old (.Channel):
                Existing channel with local data.
            new (.Channel):
                Target replacement channel to migrate data to.

        Returns:
            str list:
                Names of hooks that migrated any data for the requested channel.
        """
        hooks = list(self.hooks.values())
        if not hooks:
            return []
        results = await gather(*(hook.channel_migrate(old, new) for hook in hooks))
        return [hook.name for hook, result in zip(hooks, results) if result]

    async def process(self):
        """
        Retrieve messages from plugs, and distribute them to hooks.
        """
        if self._stream:
            raise RuntimeError("Host is already processing")
        self._stream = PlugStream()
        self._stream.add(*self.plugs.values())
        try:
            async for sent, source, primary in self._stream:
                await self._callback(sent, source, primary)
        except StopAsyncIteration:
            log.debug("Plug stream finished")
        finally:
            self._stream = None

    async def run(self):
        """
        Main entry point for running as a full application.  Opens all plugs and hooks, blocks
        (when awaited) for the duration of :meth:`process`, and closes openables during shutdown.
        """
        if self._process:
            raise RuntimeError("Host is already running")
        if not self._loaded:
            self.loaded()
        await self.open()
        try:
            self._process = ensure_future(self.process())
            await self._process
        except CancelledError:
            log.debug("Host run cancelled")
        finally:
            self._process = None
            await self.close()

    async def quit(self):
        """
        Request the running host to stop processing.  This only works if started via :meth:`run`.
        """
        if self._process:
            self._process.cancel()

    def __repr__(self):
        return "<{}: {}P {}C {}G {}R {}H{}>".format(self.__class__.__name__, len(self.plugs),
                                                    len(self.channels), len(self.groups),
                                                    len(self._resources), len(self.plain_hooks),
                                                    " running" if self.running else "")



from collections import defaultdict
from reprlib import recursive_repr


class Any:
    """
    Allow values to match from a choice of schemas.  :class:`Invalid` will be raised if none of
    them are valid for a given value.

    If no schemas are defined, this acts as a wildcard, i.e. it will match *any* value given.

    When used as a dictionary key, exactly one choice must be present.  To allow any number of keys,
    use :class:`.Optional` instead.

    Attributes:
        choices (.Schema list):
            Set of valid schemas, or an empty set to match any possible value.
    """

    __slots__ = ("choices",)

    def __init__(self, *choices):
        self.choices = list(choices)

    def __eq__(self, other):
        return (self.choices == other.choices if isinstance(other, self.__class__)
                else tuple(self.choices) == other)

    def __hash__(self):
        return hash(tuple(self.choices))

    @recursive_repr()
    def __repr__(self):
        return "<{}{}>".format(self.__class__.__name__,
                               ": {}".format(", ".join(repr(choice) for choice in self.choices))
                               if self.choices else "")


class Nullable:
    """
    Allow values tested against the inner schema to hold a null value.  Without this wrapper,
    ``None`` will not be accepted by a schema and will raise :class:`Invalid`, whereas
    ``Nullable(str)`` matches both ``"foo"`` and ``None``.

    This object evaluates equally to its inner schema.

    Attributes:
        schema (.Schema):
            Inner schema for non-null processing.
    """

    __slots__ = ("schema",)

    @classmethod
    def unwrap(cls, value):
        """
        Unpack a value, whether nullable or not.

        Args:
            value:
                Plain or :class:`.Nullable` schema value.

        Returns:
            Tuple of ``(schema, nullable)``, where :data:`nullable` is ``True`` if the source value
            was :class:`.Nullable`, ``False`` otherwise.
        """
        return (value.schema, True) if isinstance(value, cls) else (value, False)

    def __init__(self, schema):
        self.schema = schema

    def __eq__(self, other):
        return self.schema == (other.schema if isinstance(other, self.__class__) else other)

    def __hash__(self):
        return hash(self.schema)

    @recursive_repr()
    def __repr__(self):
        return "<{}: {!r}>".format(self.__class__.__name__, self.schema)


class Optional:
    """
    Allows keys matching the inner schema to not be present in the source :class:`dict`.  Without
    this wrapper, :class:`Invalid` will be raised if the corresponding key is missing.

    This object evaluates equally to its inner schema.

    Attributes:
        schema (.Schema):
            Inner schema for processing.
        default:
            Alternative value for when the key is missing.  This can be :data:`None` (the default),
            an instance of a static class (:class:`int`, :class:`float`, :class:`bool`), the
            :class:`list` or :class:`dict` constructor, or a lambda or function that produces the
            desired value (e.g. ``lambda: {"a": 1}``).

            When using ``None``, you should also mark the corresponding value as :class:`.Nullable`.
    """

    MISSING = object()

    __slots__ = ("schema", "default")

    @classmethod
    def unwrap(cls, key):
        """
        Unpack a :class:`dict` key, whether optional or not.

        Args:
            key:
                Plain or :class:`.Optional` key.

        Returns:
            Tuple of ``(schema, default)``, where ``default`` is :attr:`.Optional.MISSING` if
            :data:`key` is not an :class:`.Optional` instance.
        """
        return (key.schema, key.default) if isinstance(key, cls) else (key, cls.MISSING)

    def __init__(self, schema, default=None):
        self.schema = schema
        self.default = default

    def __eq__(self, other):
        return self.schema == (other.schema if isinstance(other, self.__class__) else other)

    def __hash__(self):
        return hash(self.schema)

    @recursive_repr()
    def __repr__(self):
        return "<{}: {!r} -> {!r}>".format(self.__class__.__name__, self.schema, self.default)


class SchemaError(Exception):
    """
    Error with the definitiion of the schema itself, raised during validation.
    """


class Invalid(Exception):
    """
    Error with input data not matching the corresponding schema.
    """


class Walker:
    """
    Base class for stepping through all nodes of a :class:`.Schema` instance.
    """

    STATIC = (int, float, bool, str)

    RECURSE = object()

    @classmethod
    def _at_path(cls, text, path):
        return "{}{}".format(text, " (path: {})".format(path) if path else "")

    @classmethod
    def _has(cls, item, objs):
        return any(item is obj for obj in objs)

    @classmethod
    def recurse(cls, obj, path, seen, *args):
        """
        Handle recursion of the schema.  By default, raises :class:`RecursionError` when a node is
        discovered again during a single call.

        If :attr:`.Walker.RECURSE` is returned, the node will be processed again.  Otherwise, the
        resulting value (including ``None``) will be used for the inner repeat of this node without
        any further processing.

        Arguments:
            obj (.Schema):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.

        Raises:
            RecursionError:
                To block looping through the same portion of the schema.
        """
        raise RecursionError(cls._at_path(repr(obj), path))

    @classmethod
    def static(cls, obj, path, seen, *args):
        """
        Handle a static object or type, as defined by :attr:`.Walker.STATIC`.

        Arguments:
            obj (int, float, bool, str):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        return obj

    @classmethod
    def nullable(cls, obj, path, seen, *args):
        """
        Handle a nullable wrapper.

        Arguments:
            obj (.Nullable):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        return cls.dispatch(Nullable.unwrap(obj)[0], path, seen, *args)

    @classmethod
    def any(cls, obj, path, seen, *args):
        """
        Handle a multiple-choice wrapper.

        Arguments:
            obj (.Nullable):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        static = tuple(choice for choice in obj.choices if choice in cls.STATIC)
        for choice in obj.choices:
            if choice is None or isinstance(choice, Nullable):
                raise SchemaError(cls._at_path("Use outer Nullable() instead of None", path))
            elif isinstance(choice, Optional):
                raise SchemaError(cls._at_path("Use Optional() outside of Any()", path))
            elif isinstance(choice, static):
                if isinstance(choice, bool) and bool not in static:
                    continue
                raise SchemaError(cls._at_path("Useless static value duplicated by type", path))
        return Any(*(cls.dispatch(item, "{}:any({})".format(path, pos), seen, *args)
                     for pos, item in enumerate(obj.choices)))

    @classmethod
    def list(cls, obj, path, seen, *args):
        """
        Handle a list.  Multiple list members are considered equivalent to an :class:`.Any`.

        Arguments:
            obj (list):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        if obj is list or not obj:
            return obj
        elif len(obj) == 1:
            return [cls.dispatch(obj[0], "{}[0]".format(path), seen, *args)]
        else:
            # [int, str] == [Any(int, str)]
            return cls.any(Any(*obj), path, seen, *args).choices

    @classmethod
    def dict(cls, obj, path, seen, *args):
        """
        Handle a dictionary.

        Arguments:
            obj (dict):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        if obj is dict or not obj:
            return dict
        for key in obj:
            if isinstance(key, Optional):
                item, default = Optional.unwrap(key)
                if isinstance(default, (list, dict)):
                    raise SchemaError(cls._at_path("Use constructor instead of {} instance"
                                                   .format(type(obj).__name__), path))
            else:
                item = key
            choices = item.choices if isinstance(item, Any) else [item]
            for choice in choices:
                if not isinstance(choice, str) and choice is not str:
                    raise SchemaError(cls._at_path("Dictionary keys must be str, not {}"
                                                   .format(type(choice).__name__), path))
        return {key: cls.dispatch(value, "{}.{}".format(path, key), seen, *args)
                for key, value in obj.items()}

    @classmethod
    def dispatch(cls, obj, path, seen, *args):
        """
        Defer to other helper methods based on the input type.

        Arguments:
            obj (.Schema):
                Schema node.
            path (str):
                Path taken through the schema from the root.
            seen (.Schema list):
                Nodes already visited in the schema.
        """
        if cls._has(obj, seen):
            recursed = cls.recurse(obj, path, seen, *args)
            if recursed is not cls.RECURSE:
                return recursed
        elif not isinstance(obj, cls.STATIC + (type,)) and obj is not None:
            seen = seen + [obj]
        if cls._has(obj, cls.STATIC) or isinstance(obj, cls.STATIC):
            return cls.static(obj, path, seen, *args)
        elif isinstance(obj, Nullable):
            return cls.nullable(obj, path, seen, *args)
        elif isinstance(obj, Any):
            return cls.any(obj, path, seen, *args)
        elif obj is list or isinstance(obj, list):
            return cls.list(obj, path, seen, *args)
        elif obj is dict or isinstance(obj, dict):
            return cls.dict(obj, path, seen, *args)
        elif isinstance(obj, Schema):
            return cls.dispatch(Schema.unwrap(obj), path, seen, *args)
        else:
            raise SchemaError(cls._at_path("Unknown type {}".format(type(obj).__name__), path))

    @classmethod
    def walk(cls, obj, *args):
        """
        Main entrypoint to the walk.

        Arguments:
            obj (.Schema):
                Schema node.

        Raises:
            SchemaError:
                When the schema is misconfigured.
        """
        return cls.dispatch(obj, "", [], *args)


class Validator(Walker):
    """
    Validation of schemas against input data.
    """

    @classmethod
    def _short(cls, obj):
        if obj is None:
            return "None"
        elif isinstance(obj, type):
            return obj.__name__
        else:
            return "{} {!r}".format(type(obj).__name__, obj)

    @classmethod
    def recurse(cls, obj, path, seen, data):
        return cls.RECURSE

    @classmethod
    def static(cls, obj, path, seen, data):
        # Don't allow the usual subclassing of ints as bools.
        if obj is int and isinstance(data, int) and not isinstance(data, bool):
            return data
        elif obj in (float, bool, str) and isinstance(data, obj):
            return data
        elif (obj == data and isinstance(obj, int) and
              not isinstance(obj, bool) and not isinstance(data, bool)):
            return data
        elif (obj == data and isinstance(obj, (float, bool, str)) and
              isinstance(data, (float, bool, str))):
            return data
        else:
            raise Invalid(cls._at_path("Expecting {} but got {}"
                                       .format(cls._short(obj), cls._short(data)), path))

    @classmethod
    def nullable(cls, obj, path, seen, data):
        if data is None:
            return data
        else:
            return super().nullable(obj, path, seen, data)

    @classmethod
    def any(cls, obj, path, seen, data):
        if not obj.choices:
            return data
        excs = []
        for pos, choice in enumerate(obj.choices):
            try:
                return cls.dispatch(choice, "{}:any({})".format(path, pos), seen, data)
            except Invalid as e:
                excs.append(e)
        else:
            # No schemas matched the data.
            raise Invalid(cls._at_path("No matches for Any()", path), *excs)

    @classmethod
    def list(cls, obj, path, seen, data):
        if not isinstance(data, list):
            raise Invalid(cls._at_path("Expecting list but got {}"
                                       .format(cls._short(data)), path))
        elif obj is list:
            return data
        elif len(obj) == 1:
            return [cls.dispatch(obj[0], "{}[{}]".format(path, pos), seen, item)
                    for pos, item in enumerate(data)]
        else:
            multi = Any(*obj)
            return [cls.any(multi, "{}[{}]".format(path, pos), seen, item)
                    for pos, item in enumerate(data)]

    @classmethod
    def dict(cls, obj, path, seen, data):
        if not isinstance(data, dict):
            raise Invalid(cls._at_path("Expecting dict but got {}"
                                       .format(cls._short(data)), path))
        elif obj is dict or not obj:
            return dict(data)
        parsed = {}
        optional = dict(Optional.unwrap(key) for key in obj if isinstance(key, Optional))
        for item in obj:
            key = Optional.unwrap(item)[0]
            if isinstance(key, Any):
                matches = [choice for choice in key.choices if choice in data]
                if len(matches) > 1:
                    raise Invalid(cls._at_path("Multiple matches for Any()", path))
                elif matches:
                    parsed[matches[0]] = cls.dispatch(obj[item], "{}.{}".format(path, matches[0]),
                                                      seen, data[matches[0]])
                elif not cls._has(key, optional):
                    raise Invalid(cls._at_path("No matches for Any()", path))
        for key in obj:
            if isinstance(key, str) and key not in data:
                if key in optional:
                    parsed[key] = optional[key]
                else:
                    raise Invalid(cls._at_path("Missing key {!r}".format(key), path))
        typed = tuple(key for key in obj if isinstance(key, type))
        fixed = {key for key in obj if not isinstance(key, type)}
        for key, value in data.items():
            here = "{}.{}".format(path, key)
            if key in fixed:
                parsed[key] = cls.dispatch(obj[key], here, seen, value)
                continue
            for match in typed:
                if isinstance(key, match):
                    parsed[key] = cls.dispatch(obj[match], here, seen, value)
                    break
            else:
                if key not in parsed:
                    # Unmatched keys are passed through without further validation.
                    parsed[key] = value
        for key in optional:
            here = "{}.{}".format(path, key)
            if key in data or isinstance(key, Any):
                continue
            # Missing but optional keys are filled in and validated.
            default = optional[key]
            if callable(default):
                default = default()
            parsed[key] = cls.dispatch(obj[key], here, seen, default)
        return parsed

    @classmethod
    def dispatch(cls, obj, path, seen, data):
        if isinstance(data, type):
            raise Invalid(cls._at_path("Expecting instance but got {} type"
                                       .format(data.__name__), path))
        else:
            return super().dispatch(obj, path, seen, data)

    @classmethod
    def walk(cls, obj, data):
        """
        Validate the given data against a schema.

        Args:
            obj (.Schema):
                Description of the data format.
            data:
                Input data to validate.

        Raises:
            Invalid:
                When a key or value doesn't match the accepted type for that field.

        Returns:
            Parsed data with optional values filled in.
        """
        return super().walk(obj, data)


class JSONSchema(Walker):
    """
    Generator of `JSON Schema <https://json-schema.org>`_ objects, suitable for external validation
    of schemas in JSON.
    """

    TYPES = {int: "number", float: "number", bool: "boolean", str: "string"}

    @classmethod
    def _make_anyof(cls, choices):
        types = []
        anys = []
        consts = defaultdict(set)
        for choice in choices:
            if not choice:
                return {}
            elif isinstance(choice, dict) and "type" in choice:
                if len(choice) == 1:
                    types.append(choice["type"])
                elif len(choice) == 2 and "const" in choice:
                    consts[choice["type"]].add(choice["const"])
                elif len(choice) == 2 and "enum" in choice:
                    consts[choice["type"]].update(choice["enum"])
                else:
                    anys.append(choice)
            else:
                anys.append(choice)
        for type_, values in consts.items():
            if type_ in types or not values:
                continue
            elif len(values) == 1:
                anys.append({"type": type_, "const": next(iter(values))})
            else:
                anys.append({"type": type_, "enum": list(values)})
        if len(types) == 1:
            anys.append({"type": types[0]})
        elif types:
            anys.append({"type": types})
        if len(anys) == 1:
            return anys[0]
        elif anys:
            return {"anyOf": anys}
        else:
            return {}

    @classmethod
    def static(cls, obj, path, seen):
        if isinstance(obj, type):
            return {"type": cls.TYPES[obj]}
        else:
            return {"type": cls.TYPES[type(obj)], "const": obj}

    @classmethod
    def nullable(cls, obj, path, seen):
        return cls._make_anyof([{"type": "null"}, super().nullable(obj, path, seen)])

    @classmethod
    def any(cls, obj, path, seen):
        return cls._make_anyof(super().any(obj, path, seen).choices)

    @classmethod
    def list(cls, obj, path, seen):
        root = {"type": "array"}
        if obj is list or not obj:
            return root
        elif len(obj) > 1:
            root["items"] = cls._make_anyof(super().list(obj, path, seen))
        else:
            root["items"] = cls.dispatch(obj[0], path, seen)
        return root

    @classmethod
    def dict(cls, obj, path, seen):
        root = {"type": "object"}
        if obj is dict or not obj:
            return root
        optional = dict(Optional.unwrap(key) for key in obj if isinstance(key, Optional))
        fixed = {key for key in obj if not isinstance(key, type)}
        if fixed:
            root["properties"] = {}
            for key in fixed:
                if isinstance(key, Any):
                    raise SchemaError(cls._at_path("Any() in dictionary keys not supported", path))
                item = Optional.unwrap(key)[0]
                here = "{}.{}".format(path, item)
                prop = cls.dispatch(obj[key], here, seen)
                root["properties"][item] = prop
                if key in optional:
                    default = optional[key]
                    prop["default"] = default() if callable(default) else default
            required = [key for key in fixed if key not in optional]
            if required:
                root["required"] = required
        typed = tuple(key for key in obj if isinstance(key, type))
        if typed:
            root["additonalItems"] = cls.any(Any(*(obj[key] for key in typed)), path, seen)
        return root

    @classmethod
    def walk(cls, obj):
        """
        Convert a schema structure into a `JSON Schema <https://json-schema.org>`_ representation.

        Args:
            schema (.Schema):
                Input schema or structure.

        Returns:
            dict:
                Equivalent JSON Schema data.
        """
        schema = super().walk(obj)
        schema["$schema"] = "http://json-schema.org/schema#"
        return schema


class Schema:
    """
    Validate JSON-like Python structures and provide defaults:

    .. code-block:: python

        config = Schema({
            "flag": bool,
            "numbers": [int],
            "nullable": Nullable(str),
            "nested": {
                Optional("defaulted", 1): int,
                "multiple": Any(int, str)
            }
        })

        validated = config(data)

    Pass a structure representing the expected data format to the constructor, along with an
    optional :data:`base` to extend from, then validate some given data against the schema by
    calling the instance -- see :class:`Validator`.

    Attributes:
        raw:
            Root schema item, including any base items.
        json (dict):
            `JSON Schema <https://json-schema.org>`_ data corresponding to this schema -- see
            :class:`JSONSchema`.
    """

    STATIC = (int, float, bool, str)
    JSON_TYPES = {int: "number", float: "number", bool: "boolean", str: "string"}

    __slots__ = ("raw",)

    @classmethod
    def unwrap(cls, schema):
        return schema.raw if isinstance(schema, cls) else schema

    def __init__(self, raw, base=None):
        raw = Schema.unwrap(raw)
        if base is not None:
            base = Schema.unwrap(base)
            if not isinstance(raw, dict) or not isinstance(base, dict):
                raise SchemaError("Input and base schemas must both be dicts")
            merged = dict(base)
            merged.update(raw)
            raw.update(merged)
        self.raw = raw

    def __call__(self, data):
        return Validator.walk(self, data)

    @property
    def json(self):
        return JSONSchema.walk(self)

    def __repr__(self):
        return "<{}: {!r}>".format(self.__class__.__name__, self.raw)



"""
Run a webserver for other plugs or hooks to accept incoming HTTP requests.

Config:
    host (str):
        Hostname or IP address to bind to.
    port (int):
        Port number to bind to.

As the server is unauthenticated, you will typically want to bind it to localhost, and proxy it
behind a full webserver like nginx to separate out routes, lock down access and so on.

.. note::
    For templating via Jinja2, both :mod:`jinja2` and :mod:`aiohttp_jinja2` must be installed.
"""

from functools import wraps
import json
import logging
import os.path

from aiohttp import web

import immp


try:
    import aiohttp_jinja2
    from jinja2 import PackageLoader, PrefixLoader
except ImportError:
    aiohttp_jinja2 = PackageLoader = PrefixLoader = None


log = logging.getLogger(__name__)


class WebContext:
    """
    Abstraction from :mod:`aiohttp` to provide routing and templating to other hooks.

    Attributes:
        hook (.WebHook):
            Parent hook instance providing the webserver.
        prefix (str):
            URL prefix acting as the base path.
        module (str):
            Dotted module name of the Python module using this context (``__name__``).
        path (str):
            Base path of the module (``os.path.dirname(__file__)``), needed for static routes.
        env (dict):
            Additional variables to make available in the Jinja context.  The default context is:

            * :data:`immp`, the top-level :mod:`immp` module
            * :data:`host`, the running :class:`.Host` instance
            * :data:`ctx`, this :class:`.WebContext` instance
            * :data:`request`, the current :class:`aiohttp.web.Request` instance
    """

    def __init__(self, hook, prefix, module, path=None, env=None):
        self.hook = hook
        self.prefix = prefix
        self.module = module
        self.path = path
        self.env = {"ctx": self}
        if env:
            self.env.update(env)
        self._routes = {}
        self.hook.add_loader(self.module)

    def route(self, method, route, fn, template=None, name=None):
        """
        Add a new route to the webserver.

        Args:
            method (str):
                HTTP verb for the route (``GET``, ``POST`` etc.).
            route (str):
                URL pattern to match.
            fn (function):
                Callable to render the response, accepting a :class:`aiohttp.web.Request` argument.
            template (str):
                Optional template path, relative to the module path.  If specified, the view
                callable should return a context :class:`dict` which is passed to the template.
            name (str):
                Custom name for the route, defaulting to the function name if not specified.
        """
        name = name or fn.__name__
        if name in self._routes:
            raise KeyError(name)
        if template:
            fn = self._jinja(fn, template)
        route = self.hook.add_route(method, "{}/{}".format(self.prefix, route), fn,
                                    name="{}:{}".format(self.module, name))
        self._routes[name] = route

    def static(self, route, path, name=None):
        """
        Add a new route to the webserver.

        Args:
            route (str):
                URL pattern to match.
            path (str):
                Filesystem location relative to the base path.
            name (str):
                Custom name for the route, defaulting to the function name if not specified.
        """
        name = name or path
        if name in self._routes:
            raise KeyError(name)
        route = self.hook.add_static("{}/{}".format(self.prefix, route),
                                     os.path.join(self.path, path),
                                     name="{}:{}".format(self.module, name))
        self._routes[name] = route

    def _jinja(self, fn, path):
        @wraps(fn)
        async def inner(request):
            env = dict(self.env)
            env["request"] = request
            env.update(await fn(request))
            return env
        if not aiohttp_jinja2:
            raise immp.HookError("Templating requires Jinja2 and aiohttp_jinja2")
        outer = aiohttp_jinja2.template("{}/{}".format(self.module, path))
        return outer(inner)

    def url_for(self, name_, **kwargs):
        """
        Generate an absolute URL for the named route.

        Args:
            name (str):
                Route name, either the function name or the custom name given during registration.

        Returns:
            str:
                Relative URL to the corresponding page.
        """
        return self._routes[name_].url_for(**kwargs)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.prefix)


class WebHook(immp.ResourceHook):
    """
    Hook that provides a generic webserver, which other hooks can bind routes to.

    Attributes:
        app (aiohttp.web.Application):
            Web application instance, used to add new routes.
    """

    schema = immp.Schema({"host": str, "port": int})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.app = web.Application()
        if aiohttp_jinja2:
            # Empty mapping by default, other hooks can add to this via add_loader().
            self._loader = PrefixLoader({})
            self._jinja = aiohttp_jinja2.setup(self.app, loader=self._loader)
            self._jinja.filters["json"] = json.dumps
            self._jinja.globals["immp"] = immp
            self._jinja.globals["host"] = self.host
        self._runner = web.AppRunner(self.app)
        self._site = None
        self._contexts = {}

    def context(self, prefix, module, path=None, env=None):
        """
        Retrieve a context for the current module.

        Args:
            prefix (str):
                URL prefix acting as the base path.
            module (str):
                Dotted module name of the Python module using this context.  Callers should use
                :data:`__name__` from the root of their module.
            path (str):
                Base path of the module, needed for static routes.  Callers should use
                ``os.path.dirname(__file__)`` from the root of their module.
            env (dict):
                Additional variables to make available in the Jinja context.  See
                :attr:`.WebContext.env` for details.

        Returns:
            .WebContext:
                Linked context instance for that module.
        """
        self._contexts[module] = WebContext(self, prefix, module, path, env)
        return self._contexts[module]

    def add_loader(self, module):
        """
        Register a Jinja2 package loader for the given module.

        Args:
            module (str):
                Module name to register.
        """
        if not aiohttp_jinja2:
            raise immp.HookError("Loaders require Jinja2 and aiohttp_jinja2")
        self._loader.mapping[module] = PackageLoader(module)

    def add_route(self, *args, **kwargs):
        """
        Equivalent to :meth:`aiohttp.web.UrlDispatcher.add_route`.
        """
        return self.app.router.add_route(*args, **kwargs)

    def add_static(self, *args, **kwargs):
        """
        Equivalent to :meth:`aiohttp.web.UrlDispatcher.add_static`.
        """
        return self.app.router.add_static(*args, **kwargs)

    async def start(self):
        log.debug("Starting server on %s:%d", self.config["host"], self.config["port"])
        await self._runner.setup()
        self._site = web.TCPSite(self._runner, self.config["host"], self.config["port"])
        await self._site.start()

    async def stop(self):
        if self._site:
            log.debug("Stopping server")
            await self._runner.cleanup()
            self._site = None



"""
Bridge multiple channels into a single unified conversation, or relay messages from one channel to
one or more others.

Sync
~~~~

Config:
    channels ((str, str list) dict):
        Mapping from virtual channel names to lists of channel names to bridge.
    plug (str):
        Name of a virtual plug to register for this sync.
    joins (bool):
        Whether to sync join and part messages across the bridge.
    renames (bool):
        Whether to sync channel title changes across the bridge.
    identities (str):
        Name of a registered :class:`.IdentityProvider` to provide unified names across networks.

        If enabled, this will rewrite mentions for users identified in both the source and any sync
        target channels, to use their platform-native identity.
    reset-author (bool):
        ``True`` to create and attach a new user with just a name, ``False`` (default) to clone and
        modify the existing user (thus keeping username, avatar etc.).
    name-format (str):
        Template to use for replacing real names on synced messages, parsed by :mod:`jinja2`.  If
        not set but the user is identified, it defaults to ``<real name> (<identity name>)``.

        Context variables:
            user (.User):
                Message author, may be ``None`` for system messages or when ``reset-author`` is set.
            identity (.IdentityGroup):
                Connected identity, or ``None`` if no link or ``identities`` isn't set.
            title (str):
                Channel title.
    strip-name-emoji (bool):
        ``True`` to remove emoji characters from message authors' real names.

Commands:
    sync-members:
        List all members of the current conversation, across all channels.
    sync-list:
        List all channels connected to this conversation.

When a message is received from any of the listed channels, a copy is pushed to all other channels
participating in the bridge.

If ``plug`` is specified, a virtual plug is registered under that name, with a channel for each
defined bridge.  Other hooks may reference these channels, to work with all channels in that sync
as one.  This allows them to listen to a unified stream of messages, or push new messages to all
synced channels.

Forward
~~~~~~~

Config:
    channels ((str, str list) dict):
        Mapping from source channel names to lists of channel names to forward to.
    users (str list):
        Whitelist of user IDs to accept source messages from.  If set, messages from anyone else in
        the source channel will be ignored.
    joins (bool):
        Whether to forward join and part messages.
    renames (bool):
        Whether to forward channel title changes.
    identities (str):
        Name of a registered :class:`.IdentityProvider` to provide unified names across networks.

        If enabled, this will rewrite mentions for users identified in both the source and any sync
        target channels, to use their platform-native identity.
    reset-author (bool):
        ``True`` to create and attach a new user with just a name, ``False`` (default) to clone and
        modify the existing user (thus keeping username, avatar etc.).
    name-format (str):
        Template to use for replacing real names on synced messages, parsed by :mod:`jinja2`.  If
        not set but the user is identified, it defaults to ``<real name> (<identity name>)``.

        Context variables:
            user (.User):
                Message author, may be ``None`` for system messages or when ``reset-author`` is set.
            identity (.IdentityGroup):
                Connected identity, or ``None`` if no link or ``identities`` isn't set.
            title (str):
                Channel title.
    strip-name-emoji (bool):
        ``True`` to remove emoji characters from message authors' real names.

When a message is received in a configured source channel, a copy is pushed to all downstream
channels.  Unlike a sync, this is a one-direction copy, useful for announcements or alerts.

.. note::
    Use of ``name-format`` requires the `Jinja2 <http://jinja.pocoo.org>`_ Python module.  Use of
    ``strip-name-emoji`` requires the **emoji** Python module.
"""

from asyncio import BoundedSemaphore, gather
from collections import defaultdict
from copy import copy
from itertools import chain
import logging

from peewee import CharField

import immp
from immp.hook.command import command
from immp.hook.database import BaseModel, DatabaseHook
from immp.hook.identity import IdentityProvider


try:
    from jinja2 import Template
except ImportError:
    Template = None

try:
    from emoji import get_emoji_regexp
except ImportError:
    EMOJI_REGEX = None
else:
    EMOJI_REGEX = get_emoji_regexp()


log = logging.getLogger(__name__)


class SyncBackRef(BaseModel):
    """
    One of a set of references, each pointing to a representation of a source message.

    Attributes:
        key (str):
            Shared ID for each message in the same set.
        network (str):
            Network identifier of the plug for this message.
        channel (str):
            Origin channel of the referenced message.
        message (str):
            Message ID as generated by the plug.
    """

    key = CharField()
    network = CharField()
    channel = CharField()
    message = CharField()

    class Meta:
        # Unique constraint for each message.
        indexes = ((("network", "channel", "message"), True),)

    @classmethod
    def map_from_sent(cls, sent):
        """
        Take a :class:`.Receipt` and attempt to resolve it to a key from a previously synced
        reference, then defer to :meth:`map_from_key` to make the mapping with the key.

        Args:
            sent (.Receipt):
                Referenced message to lookup.

        Returns:
            ((str, str), .SyncBackRef) dict:
                Generated reference mapping.
        """
        alias = cls.alias()
        key = alias.select(alias.key).where(alias.network == sent.channel.plug.network_id,
                                            alias.channel == sent.channel.source,
                                            alias.message == str(sent.id))
        return cls.map_from_key(key)

    @classmethod
    def map_from_key(cls, key):
        """
        For a given key, fetch all messages references synced from the corresponding source, and
        group them by network and channel.

        Args:
            key (str):
                Synced message identifier.

        Returns:
            ((str, str), .SyncBackRef) dict:
                Generated reference mapping.
        """
        backrefs = cls.select().where(cls.key == key)
        if not backrefs:
            raise KeyError
        mapped = defaultdict(list)
        for backref in backrefs:
            mapped[(backref.network, backref.channel)].append(backref)
        return key, mapped

    def __repr__(self):
        return "<{}: #{} {} {} @ {}/{}>".format(self.__class__.__name__, self.id, self.key,
                                                self.message, self.network, self.channel)


class SyncRef:
    """
    Representation of a single synced message.

    Attributes:
        key (str):
            Unique synced message identifier, used by :class:`.SyncPlug` when yielding messages.
        ids ((.Channel, str) dict):
            Mapping from :class:`.Channel` to a list of echoed message IDs.
        revisions ((.Channel, (str, set) dict) dict):
            Mapping from :class:`.Channel` to message ID to synced revisions of that message.
        source (.Message):
            Original copy of the source message, if we have it.
    """

    next_key = immp.IDGen()

    @classmethod
    def from_backref_map(cls, key, mapped, host):
        """
        Take a mapping generated in :meth:`.SyncBackRef.map_from_key` and produce a local reference
        suitable for the memory cache.

        Args:
            key (str):
                Synced message identifier.
            mapped (((str, str), .SyncBackRef list) dict):
                Generated reference mapping.
            host (.Host):
                Parent host instance, needed to resolve network IDs to plugs.

        Returns:
            .SyncRef:
                Newly created reference.
        """
        ids = {}
        for (network, source), synced in mapped.items():
            for plug in host.plugs.values():
                if plug.network_id == network:
                    ids[immp.Channel(plug, source)] = [backref.message for backref in synced]
        return cls(ids, key=key)

    def __init__(self, ids, *, key=None, source=None, origin=None):
        self.key = key or self.next_key()
        self.ids = defaultdict(list, ids)
        self.revisions = defaultdict(lambda: defaultdict(set))
        self.source = source
        if origin:
            self.ids[origin.channel].append(origin.id)
            self.revision(origin)

    def revision(self, sent):
        """
        Log a new revision of a message.

        Args:
            sent (.Receipt):
                Updated message relating to a previously synced message.

        Returns:
            bool:
                ``True`` if this is an edit (i.e. we've already seen a base revision for this
                message) and needs syncing to other channels.
        """
        self.revisions[sent.channel][sent.id].add(sent.revision)
        return len(self.revisions[sent.channel][sent.id]) > 1

    def __repr__(self):
        return "<{}: #{} x{}{}>".format(self.__class__.__name__, self.key, len(self.ids),
                                        " {}".format(repr(self.source)) if self.source else "")


class SyncCache:
    """
    Synced message cache manager, using both in-memory and database-based caches.

    This class has :class:`dict`-like access, using either :class:`.Receipt` objects or
    :class:`.SyncPlug` message IDs that map to :class:`.SyncRef` keys.
    """

    def __init__(self, hook):
        self._hook = hook
        self._cache = {}
        # Reverse mapping, Channel -> ID -> SyncRef.
        self._lookup = defaultdict(dict)

    def add(self, ref, back=False):
        """
        Add a :class:`.SyncRef` to the cache.  This will also commit a new :class:`.SyncBackRef` to
        the database if configured.

        Args:
            ref (.SyncRef):
                Newly synced message to store.
            back (bool):
                ``True`` if sourced from a :class:`.SyncBackRef`, and therefore doesn't need
                committing back to the database.

        Returns:
            .SyncRef:
                The same ref, useful for shorthand add-then-return.
        """
        self._cache[ref.key] = ref
        data = []
        for channel, ids in ref.ids.items():
            for id_ in ids:
                self._lookup[channel][id_] = ref.key
                data.append({"key": ref.key, "network": channel.plug.network_id,
                             "channel": channel.source, "message": id_})
        if self._hook.db and not back:
            SyncBackRef.insert_many(data).on_conflict("ignore").execute()
        return ref

    def __getitem__(self, key):
        if isinstance(key, immp.Receipt) and key.channel.plug == self._hook.plug:
            # Message from the sync channel itself, so just look the key up directly.
            key = key.id
        if isinstance(key, immp.Receipt):
            try:
                # If in the local cache, the message already passed through sync in this session.
                # Use the existing cache entry as-is (entry in _lookup <=> entry in _cache).
                return self._cache[self._lookup[key.channel][key.id]]
            except KeyError:
                if not self._hook.db:
                    raise
            # Not cached locally, but the database is configured: check there for a reference,
            # build a new SyncRef with an empty source, and cache it.
            key, mapped = SyncBackRef.map_from_sent(key)
            return self.add(SyncRef.from_backref_map(key, mapped, self._hook.host), True)
        elif isinstance(key, str):
            # As above, check the local cache directly first.
            try:
                return self._cache[key]
            except KeyError:
                if not self._hook.db:
                    raise
            # Now check the database for the key.
            _, mapped = SyncBackRef.map_from_key(key)
            return self.add(SyncRef.from_backref_map(key, mapped, self._hook.host), True)
        else:
            raise TypeError(key)

    def __repr__(self):
        return "<{}: {} refs>".format(self.__class__.__name__, len(self._cache))


class SyncPlug(immp.Plug):
    """
    Virtual plug that allows sending external messages to a synced conversation.
    """

    schema = None

    network_name = "Sync"

    @property
    def network_id(self):
        return "sync:{}".format(self.name)

    def __init__(self, name, hook, host):
        super().__init__(name, {}, host, virtual=True)
        self._hook = hook

    @classmethod
    def any_sync(cls, host, channel):
        """
        Produce a synced channel for the given source, searching across all :class:`.SyncPlug`
        instances running on the host.

        Args:
            host (.Host):
                Controlling host instance.
            channel (.Channel):
                Original channel to lookup.

        Returns:
            .Channel:
                Sync channel containing the given channel as a source, or ``None`` if not synced.
        """
        synced = [plug.sync_for(channel) for plug in host.plugs.values() if isinstance(plug, cls)]
        try:
            return next(filter(None, synced))
        except StopIteration:
            return None

    def sync_for(self, channel):
        """
        Produce a synced channel for the given source.

        Args:
            channel (.Channel):
                Original channel to lookup.

        Returns:
            .Channel:
                Sync channel containing the given channel as a source, or ``None`` if not synced.
        """
        for label, synced in self._hook.channels.items():
            if channel in synced:
                return immp.Channel(self, label)
        return None

    def in_sync(self, channel):
        """
        Retrieve the list of member channels for a given sync channel.

        Args:
            channel (.Channel):
                Virtual channel belonging to this sync instance.

        Returns:
            .Channel list:
                Channels participating in this sync.
        """
        return self._hook.channels[channel.source] if channel.plug is self else []

    async def public_channels(self):
        return [immp.Channel(self, name) for name in self._hook.channels]

    async def private_channels(self):
        return []

    async def channel_is_private(self, channel):
        return False if channel.source in self._hook.config["channels"] else None

    async def channel_title(self, channel):
        return channel.source

    async def channel_members(self, channel):
        if channel.source not in self._hook.config["channels"]:
            return None
        tasks = (synced.members() for synced in self._hook.channels[channel.source])
        members = set(chain(*(await gather(*tasks))))
        return list(sorted(members, key=lambda u: (u.plug.name, u.username or u.real_name)))

    async def put(self, channel, msg):
        if channel.source in self._hook.config["channels"]:
            ref = await self._hook.send(channel.source, msg)
            return [ref.key]
        else:
            raise immp.PlugError("Send to unknown sync channel: {}".format(repr(channel)))

    async def delete(self, sent):
        await self._hook.delete(self._hook._cache[sent.id])


class _SyncHookBase(immp.Hook):

    schema = immp.Schema({"channels": {str: [str]},
                          immp.Optional("joins", False): bool,
                          immp.Optional("renames", False): bool,
                          immp.Optional("identities"): immp.Nullable(str),
                          immp.Optional("reset-author", False): bool,
                          immp.Optional("name-format"): immp.Nullable(str),
                          immp.Optional("strip-name-emoji", False): bool})

    _identities = immp.ConfigProperty(IdentityProvider)

    def _accept(self, msg):
        if not self.config["joins"] and (msg.joined or msg.left):
            log.debug("Not syncing join/part message: %r", msg.id)
            return False
        if not self.config["renames"] and msg.title:
            log.debug("Not syncing rename message: %r", msg.id)
            return False
        return True

    async def _replace_identity_mentions(self, msg, channel):
        # Replace mentions for identified users in the target channel.
        if not msg.text:
            return
        msg.text = msg.text.clone()
        for segment in msg.text:
            if not segment.mention or segment.mention.plug == channel.plug:
                # No mention or already matches plug, nothing to do.
                continue
            identity = None
            if self.config["identities"]:
                try:
                    identity = await self._identities.identity_from_user(segment.mention)
                except Exception as e:
                    log.warning("Failed to retrieve identity information for %r",
                                segment.mention, exc_info=e)
            replaced = False
            if identity:
                # Try to find an identity corresponding to the target plug.
                for user in identity.links:
                    if user.plug == channel.plug:
                        log.debug("Replacing mention: %r -> %r", segment.mention, user)
                        segment.mention = user
                        replaced = True
                        break
            if not replaced:
                # Fallback case: replace mention with a link to the user's profile.
                if segment.mention.link:
                    log.debug("Adding fallback mention link: %r -> %r",
                              segment.mention, segment.mention.link)
                    segment.link = segment.mention.link
                else:
                    log.debug("Removing foreign mention: %r", segment.mention)
                segment.mention = None

    async def _replace_name(self, msg):
        # Use name-format or identities to render a suitable author real name.
        name = user = identity = None
        if not self.config["reset-author"]:
            user = msg.user
        if self._identities:
            try:
                identity = await self._identities.identity_from_user(msg.user)
            except Exception as e:
                log.warning("Failed to retrieve identity information for %r", msg.user,
                            exc_info=e)
        if self.config["name-format"]:
            if not Template:
                raise immp.PlugError("'jinja2' module not installed")
            title = await msg.channel.title() if isinstance(msg, immp.Receipt) else None
            context = {"user": msg.user, "identity": identity, "channel": title}
            name = Template(self.config["name-format"]).render(**context)
            if not name and self.config["reset-author"]:
                msg.user = None
        elif self.config["reset-author"]:
            msg.user = None
        elif identity:
            name = "{} ({})".format(msg.user.real_name or msg.user.username, identity.name)
        if not name:
            return
        if self.config["strip-name-emoji"]:
            if not EMOJI_REGEX:
                raise immp.PlugError("'emoji' module not installed")
            name = EMOJI_REGEX.sub("", name).strip()
        if user:
            user = copy(user)
            user.real_name = name
        else:
            user = immp.User(real_name=name)
        msg.user = user

    async def _send(self, channel, msg):
        try:
            ids = await channel.send(msg)
            log.debug("Synced IDs in %r: %r", channel, ids)
            return (channel, ids)
        except Exception:
            log.exception("Failed to relay message to channel: %r", channel)
            return (channel, [])


class SyncHook(_SyncHookBase):
    """
    Hook to propagate messages between two or more channels.

    Attributes:
        plug (.SyncPlug):
            Virtual plug for this sync, if configured.
    """

    schema = immp.Schema({immp.Optional("joins", True): bool,
                          immp.Optional("renames", True): bool,
                          immp.Optional("plug"): immp.Nullable(str)}, _SyncHookBase.schema)

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.db = None
        # Message cache, stores IDs of all synced messages by channel.
        self._cache = SyncCache(self)
        # Hook lock, to put a hold on retrieving messages whilst a send is in progress.
        self._lock = BoundedSemaphore()
        # Add a virtual plug to the host, for external subscribers.
        if self.config["plug"]:
            log.debug("Creating virtual plug: %r", self.config["plug"])
            self.plug = SyncPlug(self.config["plug"], self, host)
            host.add_plug(self.plug)
            for label in self.config["channels"]:
                host.add_channel(label, immp.Channel(self.plug, label))
        else:
            self.plug = None

    @property
    def channels(self):
        try:
            return {virtual: [self.host.channels[label] for label in labels]
                    for virtual, labels in self.config["channels"].items()}
        except KeyError as e:
            raise immp.ConfigError("No channel {} on host".format(repr(e.args[0]))) from None

    def label_for_channel(self, channel):
        labels = []
        for label, channels in self.channels.items():
            if channel in channels:
                labels.append(label)
        if not labels:
            raise immp.ConfigError("Channel {} not bridged".format(repr(channel)))
        elif len(labels) > 1:
            raise immp.ConfigError("Channel {} defined more than once".format(repr(channel)))
        else:
            return labels[0]

    async def start(self):
        try:
            self.db = self.host.resources[DatabaseHook].db
        except KeyError:
            pass
        else:
            self.db.create_tables([SyncBackRef], safe=True)

    async def stop(self):
        self.db = None

    def _test(self, channel, user):
        return any(channel in channels for channels in self.channels.values())

    @command("sync-members", test=_test)
    async def members(self, msg):
        """
        List all members of the current conversation, across all channels.
        """
        members = defaultdict(list)
        missing = False
        for synced in self.channels[msg.channel.source]:
            local = (await synced.members())
            if local:
                members[synced.plug.network_name] += local
            else:
                missing = True
        if not members:
            return
        text = immp.RichText([immp.Segment("Members of this conversation:")])
        for network in sorted(members):
            text.append(immp.Segment("\n{}".format(network), bold=True))
            for member in sorted(members[network],
                                 key=lambda member: member.real_name or member.username):
                name = member.real_name or member.username
                text.append(immp.Segment("\n"))
                if member.link:
                    text.append(immp.Segment(name, link=member.link))
                elif member.real_name and member.username:
                    text.append(immp.Segment("{} [{}]".format(name, member.username)))
                else:
                    text.append(immp.Segment(name))
        if missing:
            text.append(immp.Segment("\n"),
                        immp.Segment("(list may be incomplete)"))
        await msg.channel.send(immp.Message(user=immp.User(real_name="Sync"), text=text))

    @command("sync-list", test=_test)
    async def list(self, msg):
        """
        List all channels connected to this conversation.
        """
        text = immp.RichText([immp.Segment("Channels in this sync:")])
        for synced in self.channels[msg.channel.source]:
            text.append(immp.Segment("\n{}".format(synced.plug.network_name)))
            title = await synced.title()
            if title:
                text.append(immp.Segment(": {}".format(title)))
        await msg.channel.send(immp.Message(user=immp.User(real_name="Sync"), text=text))

    async def send(self, label, msg, origin=None):
        """
        Send a message to all channels in this sync.

        Args:
            label (str):
                Bridge that defines the underlying synced channels to send to.
            msg (.Message):
                External message to push.
            origin (.Receipt):
                Raw message that triggered this sync; if set and part of the sync, it will be
                skipped (used to avoid retransmitting a message we just received).
        """
        # Note that `origin` corresponds to the enriched message (i.e. `sent` in on_receive()),
        # and `msg` refers to the canonical copy (i.e. `source`).
        queue = []
        # Just like with plugs, when sending a new (external) message to all channels in a sync, we
        # need to wait for all plugs to complete before processing further messages.
        clone = copy(msg)
        await self._replace_name(clone)
        async with self._lock:
            for synced in self.channels[label]:
                if origin and synced == origin.channel:
                    continue
                local = copy(clone)
                await self._replace_identity_mentions(local, synced)
                queue.append(self._send(synced, local))
            # Send all the messages in parallel, and match the resulting IDs up by channel.
            ids = dict(await gather(*queue))
            ref = SyncRef(ids, source=msg, origin=origin)
            self._cache.add(ref)
            # Push a copy of the message to the sync channel, if running.
            if self.plug:
                sent = immp.SentMessage(id_=ref.key, channel=immp.Channel(self.plug, label),
                                        text=msg.text, user=msg.user, action=msg.action,
                                        reply_to=msg.reply_to, joined=msg.joined, left=msg.left,
                                        title=msg.title, attachments=msg.attachments, raw=msg)
                self.plug.queue(sent)
            return ref

    async def delete(self, ref, sent=None):
        queue = []
        for channel, ids in ref.ids.items():
            for id_ in ids:
                if not (sent and sent.channel == channel and sent.id == id_):
                    queue.append(immp.Receipt(id_, channel).delete())
        if queue:
            await gather(*queue)

    def _replace_msg(self, channel, msg):
        base = None
        if isinstance(msg, immp.SentMessage):
            base = immp.Message(text=msg.text, user=msg.user, action=msg.action,
                                reply_to=msg.reply_to, joined=msg.joined, left=msg.left,
                                title=msg.title, attachments=msg.attachments, raw=msg.raw)
        if not isinstance(msg, immp.Receipt):
            return msg
        try:
            # Given message was a resync of the source message from a synced channel.
            ref = self._cache[msg]
        except KeyError:
            # No match for this source, replace with the unqualified message.
            return base
        log.debug("Found reference to previously synced message: %r", ref.key)
        if ref.ids.get(channel):
            # Return a reference to the transport-native copy of the message.
            at = ref.source.at if isinstance(ref.source, immp.Receipt) else None
            best = ref.source or msg
            return immp.SentMessage(id_=ref.ids[channel][0], channel=channel, at=at,
                                    text=best.text, user=best.user, action=best.action,
                                    reply_to=best.reply_to, joined=best.joined, left=best.left,
                                    title=best.title, attachments=best.attachments, raw=best.raw)
        elif channel.plug == msg.channel.plug:
            # Same network, so let the plug handle it.
            return msg
        else:
            return base

    def _replace_all(self, channel, msg):
        msg.reply_to = self._replace_msg(channel, msg.reply_to)
        attachments = []
        for attach in msg.attachments:
            if isinstance(attach, immp.Message):
                replaced = self._replace_msg(channel, attach)
                if replaced:
                    attachments.append(replaced)
            else:
                attachments.append(attach)
        msg.attachments = attachments
        return msg

    async def before_receive(self, sent, source, primary):
        await super().before_receive(sent, source, primary)
        return self._replace_all(sent.channel, copy(sent))

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        try:
            label = self.label_for_channel(sent.channel)
        except immp.ConfigError:
            return
        async with self._lock:
            # No critical section here, just wait for any pending messages to be sent.
            pass
        try:
            ref = self._cache[sent]
        except KeyError:
            if sent.deleted:
                log.debug("Ignoring deleted message not in sync cache: %r", sent)
                return
            else:
                log.debug("Incoming message not in sync cache: %r", sent)
        else:
            if sent.deleted:
                log.debug("Incoming message is a delete, needs sync: %r", sent)
                await self.delete(ref)
                return
            elif ref.revision(sent):
                log.debug("Incoming message is an update, needs sync: %r", sent)
            else:
                log.debug("Incoming message already synced: %r", sent)
                return
        if not self._accept(source):
            return
        log.debug("Sending message to synced channel %r: %r", label, sent.id)
        await self.send(label, source, sent)

    async def before_send(self, channel, msg):
        await super().before_send(channel, msg)
        return (channel, self._replace_all(channel, copy(msg)))


class ForwardHook(_SyncHookBase):
    """
    Hook to propagate messages from a source channel to one or more destination channels.
    """

    schema = immp.Schema({immp.Optional("users"): immp.Nullable([str])}, _SyncHookBase.schema)

    @property
    def _channels(self):
        try:
            return {self.host.channels[key]: tuple(self.host.channels[label] for label in value)
                    for key, value in self.config["channels"].items()}
        except KeyError as e:
            raise immp.ConfigError("No such channel '{}'".format(repr(e.args[0])))

    async def send(self, channel, msg):
        """
        Send a message to all channels in this sync.

        Args:
            channel (.Channel):
                Source channel that defines the underlying forwarding channels to send to.
            msg (.Message):
                External message to push.
        """
        queue = []
        for synced in self._channels[channel]:
            clone = copy(msg)
            await self._replace_name(clone)
            await self._replace_identity_mentions(clone, synced)
            queue.append(self._send(synced, clone))
        # Send all the messages in parallel.
        await gather(*queue)

    def _accept(self, msg):
        if not super()._accept(msg):
            return False
        if self.config["users"] is not None:
            if not msg.user or msg.user.id not in self.config["users"]:
                log.debug("Not syncing message from non-whitelisted user: %r", msg.user.id)
                return False
        return True

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if primary and sent.channel in self._channels and self._accept(source):
            await self.send(sent.channel, source)



"""
Channel join control, extended by other hooks.

Config:
    hooks ((str, str list) dict):
        Mapping of controlling hooks to a list of channels they manage.
    exclude ((str, str list) dict):
        Mapping of plugs to user IDs who should be ignored during checks.
    joins (bool):
        ``True`` to check each join as it happens.
    startup (bool):
        ``True`` to run a full check of all named channels on load.
    passive (bool):
        ``True`` to log violations without actually following through with removals.

This hook implements its own protocol for test purposes, by rejecting all joins and members of a
channel.  To make full use of it, other hooks with support for channel access can determine if a
user satisfies membership of an external group or application.
"""

import asyncio
from collections import defaultdict
import logging

import immp


log = logging.getLogger(__name__)


class AccessPredicate:
    """
    Interface for hooks to provide channel access control from a backing source.
    """

    async def channel_access(self, channel, user):
        """
        Verify if a user is allowed access to a channel.

        Args:
            channel (.Channel):
                Target channel.
            user (.User):
                Incoming user to be verified.

        Returns:
            bool:
                ``True`` if the user is to be granted access.
        """
        raise NotImplementedError


class ChannelAccessHook(immp.Hook, AccessPredicate):
    """
    Hook for controlling membership of, and joins to, secure channels.
    """

    schema = immp.Schema({immp.Optional("hooks", dict): {str: [str]},
                          immp.Optional("exclude", dict): {str: [str]},
                          immp.Optional("joins", True): bool,
                          immp.Optional("startup", False): bool,
                          immp.Optional("passive", False): bool})

    @property
    def hooks(self):
        mapping = {}
        for name, channels in self.config["hooks"].items():
            try:
                hook = self.host.hooks[name]
            except KeyError:
                raise immp.ConfigError("Hook '{}' not registered to host".format(name))
            if not isinstance(hook, AccessPredicate):
                raise immp.HookError("Hook '{}' does not implement AccessPredicate".format(name))
            try:
                channels = tuple(self.host.channels[label] for label in channels)
            except KeyError as e:
                raise immp.ConfigError("Channel '{}' not registered to host".format(e.args[0]))
            mapping[hook] = channels
        return mapping

    @property
    def channels(self):
        inverse = defaultdict(list)
        for hook, channels in self.hooks.items():
            for channel in channels:
                inverse[channel].append(hook)
        return inverse

    async def channel_access(self, channel, user):
        # Example predicate to block all access.
        return False

    async def _predicate(self, hook, channel, user):
        if not isinstance(hook, AccessPredicate):
            raise immp.HookError("Hook '{}' does not implement AccessPredicate".format(hook.name))
        allow = await hook.channel_access(channel, user)
        if not allow:
            log.debug("Hook %r disallows %r in %r", hook.name, user, channel)
            if not self.config["passive"]:
                await channel.remove(user)
        return allow

    async def _verify(self, channel, user):
        if user.id in self.config["exclude"].get(user.plug.name, []):
            log.debug("Skipping excluded user %r in channel %r", user, channel)
            return
        elif await user.is_system():
            log.debug("Skipping system user %r in channel %r", user, channel)
            return
        try:
            hooks = self.channels[channel]
        except KeyError:
            return
        for hook in hooks:
            if not await self._predicate(hook, channel, user):
                break

    async def _startup_check(self):
        log.debug("Running startup access checks")
        for channel in self.channels:
            members = await channel.members()
            if not members:
                continue
            for user in members:
                await self._verify(channel, user)
        log.debug("Finished startup access checks")

    async def start(self):
        await super().start()
        if self.config["startup"]:
            asyncio.ensure_future(self._startup_check())

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not self.config["joins"] or not primary or sent != source or not source.joined:
            return
        for user in source.joined:
            await self._verify(sent.channel, user)



"""
Basic text request/response handler.

Config:
    channels (str list):
        List of channels to process responses in.
    responses ((str, str) dict):
        Mapping from match regex to response text.

Commands:
    ar-add <match> <response>:
        Add a new trigger / response pair.
    ar-remove <match>:
        Remove an existing trigger.

This hook will listen for messages in all given channels, for text content that matches any of the
defined regular expressions.  On a match, it will answer with the corresponding response.

Because all responses are defined in the config, you'll need to ensure it's saved when making
changes via the add/remove commands.
"""

import logging
import re

import immp
from immp.hook.command import CommandParser, command


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


log = logging.getLogger(__name__)


class AutoRespondHook(immp.Hook):
    """
    Basic text responses for given trigger words and phrases.
    """

    schema = immp.Schema({"groups": [str],
                          immp.Optional("responses", dict): {str: str}})

    group = immp.Group.MergedProperty("groups")

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.responses = self.config["responses"]
        self._sent = []

    @command("ar-add", parser=CommandParser.shlex)
    async def add(self, msg, match, response):
        """
        Add a new trigger / response pair.
        """
        text = "Updated" if match in self.responses else "Added"
        self.responses[match] = response
        await msg.channel.send(immp.Message(text="{} {}".format(TICK, text)))

    @command("ar-remove", parser=CommandParser.shlex)
    async def remove(self, msg, match):
        """
        Remove an existing trigger.
        """
        if match in self.responses:
            del self.responses[match]
            text = "{} Removed".format(TICK)
        else:
            text = "{} No such response".format(CROSS)
        await msg.channel.send(immp.Message(text=text))

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not await self.group.has_channel(sent.channel):
            return
        # Skip our own response messages.
        if (sent.channel, sent.id) in self._sent:
            return
        text = str(source.text)
        for match, response in self.responses.items():
            if re.search(match, text, re.I):
                log.debug("Matched regex %r in channel: %r", match, sent.channel)
                for id_ in await sent.channel.send(immp.Message(text=response)):
                    self._sent.append((sent.channel, id_))



from .util import Configurable, Openable, pretty_str


@pretty_str
class Hook(Configurable, Openable):
    """
    Base of all hook classes, performs any form of processing on messages from all connected
    plugs, via the provided host instance.

    Instantiation may raise :class:`.ConfigError` if the provided configuration is invalid.

    Attributes:
        virtual (bool):
            ``True`` if managed by another component (e.g. a hook that exposes plug functionality).
    """

    def __init__(self, name, config, host, virtual=False):
        super().__init__(name, config, host)
        self.virtual = virtual

    async def start(self):
        """
        Perform any setup tasks.
        """

    async def stop(self):
        """
        Perform any teardown tasks.
        """

    def on_load(self):
        """
        Perform any additional one-time setup that requires other plugs or hooks to be loaded.
        """

    async def channel_migrate(self, old, new):
        """
        Move any private data between channels on admin request.  This is intended to cover data
        keyed by channel sources and plug network identifiers.

        Args:
            old (.Channel):
                Existing channel with local data.
            new (.Channel):
                Target replacement channel to migrate data to.

        Returns:
            bool:
                ``True`` if any data was migrated for the requested channel.
        """
        return False

    async def before_send(self, channel, msg):
        """
        Modify an outgoing message before it's pushed to the network.  The ``(channel, msg)`` pair
        must be returned, so hooks may modify in-place or return a different pair.  This method is
        called for each hook, one after another.  If ``channel`` is modified, the sending will
        restart on the new channel, meaning this method will be called again for all hooks.

        Hooks may also suppress a message (e.g. if their actions caused it, but it bears no value
        to the network) by returning ``None``.

        Args:
            channel (.Channel):
                Original source of this message.
            msg (.Message):
                Raw message received from another plug.

        Returns:
            (.Channel, .Message) tuple:
                The augmented or replacement pair, or ``None`` to suppress this message.
        """
        return (channel, msg)

    async def before_receive(self, sent, source, primary):
        """
        Modify an incoming message before it's pushed to other hooks.  The ``sent`` object must be
        returned, so hooks may modify in-place or return a different object.  This method is called
        for each hook, one after another, so any time-consuming tasks should be deferred to
        :meth:`process` (which is run for all hooks in parallel).

        Hooks may also suppress a message (e.g. if their actions caused it, but it bears no value
        to the rest of the system) by returning ``None``.

        Args:
            sent (.SentMessage):
                Raw message received from another plug.
            source (.Message):
                Original message data used to generate the raw message, if sent via the plug (e.g.
                from another hook), equivalent to ``msg`` if the source is otherwise unknown.
            primary (bool):
                ``False`` for supplementary messages if the source message required multiple raw
                messages in order to represent it (e.g. messages with multiple attachments where
                the underlying network doesn't support it), otherwise ``True``.

        Returns:
            .SentMessage:
                The augmented or replacement message, or ``None`` to suppress this message.
        """
        return sent

    async def on_receive(self, sent, source, primary):
        """
        Handle an incoming message received by any plug.

        Args:
            sent (.SentMessage):
                Raw message received from another plug.
            source (.Message):
                Original message data used to generate the raw message, if sent via the plug (e.g.
                from another hook), equivalent to ``msg`` if the source is otherwise unknown.
            primary (bool):
                ``False`` for supplementary messages if the source message required multiple raw
                messages in order to represent it (e.g. messages with multiple attachments where
                the underlying network doesn't support it), otherwise ``True``.
        """

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)


class ResourceHook(Hook):
    """
    Variant of hooks that globally provide access to some resource.

    Only one of each class may be loaded, which happens before regular hooks, and such hooks are
    keyed by their class rather than a name, allowing for easier lookups.
    """



"""
Utility methods related to loading and saving config files, as used by the console entry point to
save the config on exit.  Creating and running :class:`.Host` via ``python -m immp`` will attach a
:class:`RunnerHook` to its resources.

Commands:
    run-write:
        Force a write of the live config out to the configured file.
"""

from asyncio import get_event_loop
import logging
import logging.config
import signal
import sys

import anyconfig

import immp
from immp.hook.command import CommandRole, CommandScope, command


log = logging.getLogger(__name__)


class _Schema:

    _plugs = {str: {"path": str, immp.Optional("config", dict): dict}}

    _hooks = {str: {"path": str,
                    immp.Optional("priority"): immp.Nullable(int),
                    immp.Optional("config", dict): dict}}

    _channels = {str: {"plug": str, "source": str}}

    _logging = {immp.Optional("disable_existing_loggers", False): bool}

    config = immp.Schema({immp.Optional("path", list): [str],
                          immp.Optional("plugs", dict): _plugs,
                          immp.Optional("channels", dict): _channels,
                          immp.Optional("groups", dict): {str: dict},
                          immp.Optional("hooks", dict): _hooks,
                          immp.Optional("logging"): immp.Nullable(_logging)})


def config_to_host(config, path, write):
    host = immp.Host()
    for name, spec in config["plugs"].items():
        cls = immp.resolve_import(spec["path"])
        host.add_plug(cls(name, spec["config"], host))
    for name, spec in config["channels"].items():
        plug = host.plugs[spec["plug"]]
        host.add_channel(name, immp.Channel(plug, spec["source"]))
    for name, group in config["groups"].items():
        host.add_group(immp.Group(name, group, host))
    for name, spec in config["hooks"].items():
        cls = immp.resolve_import(spec["path"])
        host.add_hook(cls(name, spec["config"], host), spec["priority"])
    try:
        host.add_hook(RunnerHook("runner", {}, host))
    except immp.ConfigError:
        # Prefer existing hook defined within the config itself.
        pass
    host.resources[RunnerHook].load(config, path, write)
    host.loaded()
    return host


def _handle_signal(signum, loop, task):
    # Gracefully accept a signal once, then revert to the default handler.
    def handler(_signum, _frame):
        log.info("Closing on signal")
        task.cancel()
        signal.signal(signum, original)
    original = signal.getsignal(signum)
    signal.signal(signum, handler)


def main(path, write=False):
    config = _Schema.config(anyconfig.load(path))
    for search in config["path"]:
        sys.path.append(search)
    if config["logging"]:
        logging.config.dictConfig(config["logging"])
    else:
        logging.basicConfig(level=logging.INFO)
        for handler in logging.root.handlers:
            handler.addFilter(immp.LocalFilter())
    log.info("Creating plugs and hooks")
    host = config_to_host(config, path, write)
    loop = get_event_loop()
    task = loop.create_task(host.run())
    for signum in (signal.SIGINT, signal.SIGTERM):
        _handle_signal(signum, loop, task)
    try:
        log.info("Starting host")
        loop.run_until_complete(task)
    finally:
        loop.close()
        if write:
            host.resources[RunnerHook].write_config()


class RunnerHook(immp.ResourceHook):
    """
    Virtual hook that handles reading and writing of config from/to a file.

    Attributes:
        writeable (bool):
            ``True`` if the file will be updated on exit, or ``False`` if being used read-only.
    """

    schema = None

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self._config = None
        self._path = None
        self.writeable = None

    def load(self, config, path, writeable):
        """
        Initialise the runner with a full config and the file path.

        Args:
            config (dict):
                Complete parsed config file content.
            path (str):
                Target config file location.
            writeable (bool):
                ``True`` if changes to the live config may be written back to the file.
        """
        self._config = config
        self._path = path
        self.writeable = writeable

    def _test(self, channel, user):
        return self.writeable

    @command("run-write", scope=CommandScope.private, role=CommandRole.admin, test=_test)
    async def write(self, msg):
        """
        Force a write of the live config out to the configured file.
        """
        self.write_config()
        await msg.channel.send(immp.Message(text="\N{WHITE HEAVY CHECK MARK} Written"))

    @staticmethod
    def _config_feature(section, name, obj, priority=None):
        if obj.virtual:
            return
        feature = {"path": "{}.{}".format(obj.__class__.__module__, obj.__class__.__name__)}
        if obj.schema and obj.config:
            feature["config"] = obj.config
        if priority:
            feature["priority"] = priority
        section[name] = feature

    @property
    def config_features(self):
        config = {"plugs": {}, "channels": {}, "groups": {}, "hooks": {}}
        for name, plug in self.host.plugs.items():
            self._config_feature(config["plugs"], name, plug)
        for name, channel in self.host.channels.items():
            if not channel.plug.virtual:
                config["channels"][name] = {"plug": channel.plug.name, "source": channel.source}
        for name, group in self.host.groups.items():
            config["groups"][name] = group.config
        priorities = {hook: priority for priority, hook in self.host._priority.items()}
        for name, hook in self.host.hooks.items():
            self._config_feature(config["hooks"], name, hook, priorities.get(hook))
        return config

    @property
    def config_full(self):
        config = self._config.copy()
        config.update(self.config_features)
        return config

    def write_config(self):
        """
        Write the live config out to the target config file, if writing is enabled.
        """
        if not self.writeable:
            raise immp.PlugError("Writing not enabled")
        log.info("Writing config file")
        anyconfig.dump(self.config_full, self._path)



"""
Connect to an IRC server.

Config:
    server:
        host (str):
            Hostname of the server.
        port (int):
            Non-SSL port of the server.
        ssl (bool):
            Whether to connect using SSL.
        password (str):
            Optional password required for the server.
    user:
        nick (str):
            Primary nick for the bot user.
        real-name (str):
            Real name, as displayed in WHO queries.
    quit (str):
        Quit message, sent as part of disconnection from the server.
    accept-invites (bool):
        ``True`` to auto-join channels when an INVITE is received.
"""

from asyncio import CancelledError, Future, Queue, ensure_future, open_connection, sleep
import logging
import re

import immp


log = logging.getLogger(__name__)


class IRCError(immp.PlugError):
    """
    Generic error from the IRC server.
    """


class IRCTryAgain(IRCError):
    """
    Rate-limited response to an IRC command (RPL_TRYAGAIN, 263).
    """


class Line:
    """
    Low-level representation of an IRC message, either sent or received.  Calling :func:`str` on a
    line will encode it suitable for transmission.

    Attributes:
        command (str):
            IRC verb or numeric.
        args (str list):
            Additional arguments for this message.
        source (str list):
            Optional source of this message.
        tags ((str, str) dict):
            Any tags attached to the message.
    """

    _format = re.compile("(?:@(?P<tags>[a-z0-9-]+(?:=[^; ]+)?(?:;[a-z0-9-]+(?:=[^; ]+)?)*) +)?"
                         "(?::(?P<source>[^ ]+) +)?(?P<command>[a-z]+|[0-9]{3})"
                         "(?P<args>(?: +[^: ][^ ]*)*)(?: +:(?P<trailing>.*))?", re.I)

    def __init__(self, command, *args, source=None, tags=None):
        self.command = command.upper()
        self.args = args
        self.source = source
        self.tags = tags

    # Conveniently, this generates timestamp identifiers of the desired format.
    next_ts = immp.IDGen()

    @classmethod
    def parse(cls, line):
        """
        Take a raw IRC line and decode it into a :class:`.Line`.

        Args:
            line (str):
                Raw IRC message.

        Returns:
            .Line:
                Parsed line.
        """
        match = cls._format.match(line)
        if not match:
            raise ValueError("Invalid line: '{}'".format(line))
        tagpart, source, command, argpart, trailing = match.groups()
        tags = {}
        args = []
        if tagpart:
            for item in tagpart.split(";"):
                key, *val = item.split("=", 1)
                tags[key] = val[0] if val else True
        if argpart:
            args = argpart.split()
        if trailing:
            args.append(trailing)
        return cls(command, *args, source=source, tags=tags)

    def __str__(self):
        line = self.command
        if self.source:
            line = ":{} {}".format(self.source, line)
        if self.tags:
            tagpart = []
            for key, value in self.tags:
                tagpart.append(key if value is True else "{}={}".format(key, value))
            line = "@{} {}".format(";".join(tagpart), line)
        if self.args:
            line = " ".join([line, *self.args[:-1], ":{}".format(self.args[-1])])
        return line

    def __repr__(self):
        return "<{}: {}{}{}>".format(self.__class__.__name__, self.command,
                                     " @ {}".format(repr(self.source)) if self.source else "",
                                     " {}".format(repr(list(self.args))) if self.args else "")


class IRCSegment(immp.Segment):
    """
    Plug-friendly representation of IRC message formatting.
    """

    @classmethod
    def to_formatted(cls, segment):
        """
        Convert a :class:`.Segment` into text formatted using IRC ASCII escape sequences.

        Args:
            segment (.Segment)
                Message segment created by another plug.

        Returns:
            str:
                Code-formatted string.
        """
        text = segment.text
        if segment.bold:
            text = "\x02{}\x02".format(text)
        if segment.italic:
            text = "\x1d{}\x1d".format(text)
        if segment.underline:
            text = "\x1f{}\x1f".format(text)
        if segment.strike:
            # Muted text by colouring it grey.  Includes a default background colour to avoid
            # accidental combinations with a literal comma in a following segment.
            text = "\x0314{}\x0399,99".format(text)
        if segment.link and segment.text != segment.link:
            text = "{} [{}]".format(text, segment.link)
        return text


class IRCRichText(immp.RichText):
    """
    Wrapper for IRC-specific encoding of formatting.
    """

    @classmethod
    def to_formatted(cls, rich):
        """
        Convert a :class:`.RichText` into text formatted using IRC ASCII escape sequences.

        Args:
            rich (.RichText):
                Parsed rich text container.

        Returns:
            str:
                Code-formatted string.
        """
        return "".join(IRCSegment.to_formatted(segment) for segment in rich).replace("\t", " ")


class IRCMessage(immp.Message):
    """
    Message originating from IRC.
    """

    @classmethod
    async def from_line(cls, irc, line):
        """
        Convert a :class:`.Line` into a :class:`.Message`.

        Args:
            irc (.IRCPlug):
                Related plug instance that provides the line.
            line (.Line):
                Raw message line from the server.

        Returns:
            .IRCMessage:
                Parsed message object.
        """
        channel = line.args[0]
        nick = line.source.split("!", 1)[0]
        if channel == irc.config["user"]["nick"]:
            # Private messages arrive "from A to B", and should be sent "from B to A".
            channel = nick
        user = immp.User(id_=line.source, plug=irc, username=nick, raw=line)
        action = False
        joined = []
        left = []
        if line.command == "JOIN":
            text = "joined {}".format(channel)
            action = True
            joined.append(user)
        elif line.command == "PART":
            text = "left {}".format(channel)
            action = True
            left.append(user)
        elif line.command == "KICK":
            target = await irc.user_from_username(line.args[1])
            text = immp.RichText([immp.Segment("kicked "),
                                  immp.Segment(target.username, bold=True),
                                  immp.Segment(" ({})".format(line.args[2]))])
            action = True
            left.append(target)
        elif line.command == "PRIVMSG":
            text = line.args[1]
            match = re.match(r"\x01ACTION ([^\x01]*)\x01", text)
            if match:
                text = match.group(1)
                action = True
        else:
            raise NotImplementedError
        return immp.SentMessage(id_=Line.next_ts(),
                                channel=immp.Channel(irc, channel),
                                text=text,
                                user=user,
                                action=action,
                                joined=joined,
                                left=left,
                                raw=line)


class Wait:
    """
    Request-like object for sending a :class:`.Line` and waiting on a response.

    After sending, await an instance of this class to retrieve all collected lines after a success
    line is received.  On failure, a :class:`ValueError` is raised.
    """

    def __init__(self, lines, success, fail, collect):
        self.lines = lines
        self._success = success
        self._fail = fail
        self._collect = collect
        self._data = []
        self._result = Future()

    def add(self, line):
        if line.command == "263":
            self._result.set_exception(IRCTryAgain(line))
            return True
        if line.command in self._collect:
            self._data.append(line)
        if line.command in self._success:
            self._result.set_result(self._data)
            return True
        elif line.command in self._fail:
            self._result.set_exception(IRCError(line))
            return True
        else:
            return False

    def cancel(self):
        self._result.set_exception(CancelledError("Wait was cancelled"))

    def __await__(self):
        return self._result.__await__()

    def __repr__(self):
        parts = []
        for key in ("success", "fail", "collect"):
            value = getattr(self, "_{}".format(key))
            if value:
                parts.append("{} {}".format(key, "/".join(value)))
        return "<{}: {}>".format(self.__class__.__name__, ", ".join(parts))


class IRCPlug(immp.Plug):
    """
    Plug for an IRC server.
    """

    schema = immp.Schema({"server": {"host": str,
                                     "port": int,
                                     immp.Optional("ssl", False): bool,
                                     immp.Optional("password"): immp.Nullable(str)},
                          "user": {"nick": str,
                                   "real-name": str},
                          immp.Optional("quit"): immp.Nullable(str),
                          immp.Optional("accept-invites", False): bool})

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self._reader = self._writer = None
        self._prefixes = ""
        # Bot's own identifier as seen by the IRC server.
        self._source = None
        # Tracking fields for storing requested data by type.
        self._waits = Queue()
        self._current_wait = None
        # Don't yield messages for initial self-joins.
        self._joins = set()
        # Cache responses that may be rate-limited by the server.
        self._channels = self._names = None
        self._members = {}

    @property
    def network_name(self):
        return "{} IRC".format(self.config["server"]["host"])

    @property
    def network_id(self):
        return "irc:{}".format(self.config["server"]["host"])

    async def start(self):
        host = self.config["server"]["host"]
        port = self.config["server"]["port"]
        ssl = self.config["server"]["ssl"] or None
        reader, self._writer = await open_connection(host, port, ssl=ssl)
        self._reader = ensure_future(self._read_loop(reader, host, port))
        if self.config["server"]["password"]:
            self.write(Line("PASS", self.config["server"]["password"]))
        # We won't receive this until a valid nick has been set.
        await self.wait(Line("NICK", self.config["user"]["nick"]),
                        Line("USER", "immp", "0", "*", self.config["user"]["real-name"]),
                        success=("001",))
        self._source = (await self.user_from_username(self.config["user"]["nick"])).id
        for channel in self.host.channels.values():
            if channel.plug == self and channel.source.startswith("#"):
                self._joins.add(channel.source)
                await self._join(channel.source)

    async def stop(self):
        if self._current_wait:
            self._current_wait.cancel()
            self._current_wait = None
        while not self._waits.empty():
            self._waits.get_nowait().cancel()
        self._joins.clear()
        self.write(Line("QUIT", self.config["quit"] or "IMMP: stopping"))
        if self._reader:
            log.debug("Closing reader")
            self._reader.cancel()
            self._reader = None
        if self._writer:
            log.debug("Closing writer")
            await self._writer.drain()
            self._writer.close()
            self._writer = None
        self._source = None

    async def _read_loop(self, reader, host, port):
        while True:
            raw = await reader.readline()
            if not raw:
                # Connection has been closed.
                self._writer.close()
                self._reader = self._writer = None
                break
            line = Line.parse(raw.decode().rstrip("\r\n"))
            log.debug("Received line: %r", line)
            await self._handle(line)
        log.debug("Reconnecting in 3 seconds")
        await sleep(3)
        await self.start()

    async def _who(self, name):
        if name in self._members:
            return self._members[name]
        users = set()
        for line in await self.wait(Line("WHO", name), success=("315",), collect=("352",)):
            id_ = "{}!{}@{}".format(line.args[5], line.args[2], line.args[3])
            users.add(immp.User(id_=id_, plug=self, username=line.args[5], raw=line))
        if users:
            self._members[name] = users
        elif name in self._members:
            del self._members[name]
        return users

    async def _join(self, name):
        await self.wait(Line("JOIN", name), success=("366",), collect=("353",))
        await self._who(name)

    def _strip_prefix(self, name):
        return re.sub("^[{}]+".format(self._prefixes), "", name) if self._prefixes else name

    async def user_from_id(self, id_):
        nick = id_.split("!", 1)[0]
        return immp.User(id_=id_, plug=self, username=nick)

    async def user_from_username(self, username):
        for user in await self._who(username):
            if user.username == username:
                return user
        return None

    async def user_is_system(self, user):
        return user.id == self._source

    async def public_channels(self):
        try:
            self._channels = await self.wait(Line("LIST"), success=("323",), collect=("322",))
        except IRCTryAgain:
            pass
        if self._channels:
            return [immp.Channel(self, line.args[1]) for line in self._channels]
        else:
            return None

    async def private_channels(self):
        try:
            raw = await self.wait(Line("NAMES"), success=("366",), fail=("401",), collect=("353",))
        except IRCTryAgain:
            pass
        else:
            self._names = set()
            for line in raw:
                self._names.update(self._strip_prefix(name) for name in line.args[3].split())
        if self._names:
            return [immp.Channel(self, name) for name in self._names
                    if name != self.config["user"]["nick"]]
        else:
            return None

    async def channel_for_user(self, user):
        return immp.Channel(self, user.username)

    async def channel_is_private(self, channel):
        return bool(await self.user_from_username(channel.source))

    async def channel_title(self, channel):
        return channel.source

    async def channel_members(self, channel):
        members = list(await self._who(channel.source))
        if await channel.is_private() and members[0].id != self._source:
            members.append(await self.user_from_id(self._source))
        return members

    async def channel_invite(self, channel, user):
        self.write(Line("INVITE", user.username, channel.source))

    async def channel_remove(self, channel, user):
        self.write(Line("KICK", channel.source, user.username))

    def _sync_wait(self):
        if self._current_wait or self._waits.empty():
            return
        wait = self._waits.get_nowait()
        self._current_wait = wait
        self.write(*wait.lines)

    async def _handle(self, line):
        self._sync_wait()
        if self._current_wait and self._current_wait.add(line):
            log.debug("Completing wait: %r", self._current_wait)
            self._current_wait = None
            self._sync_wait()
        if line.command == "PING":
            self.write(Line("PONG", *line.args))
        elif line.command in ("JOIN", "PART", "KICK", "PRIVMSG"):
            sent = await IRCMessage.from_line(self, line)
            if sent.joined and sent.joined[0].username == self._source.split("!", 1)[0]:
                if sent.channel.source in self._joins:
                    self._joins.remove(sent.channel.source)
                    return
            self.queue(sent)
            if line.command == "JOIN":
                log.debug("Adding %s to %s members", sent.joined[0].username, sent.channel.source)
                self._members[sent.channel.source].add(sent.joined[0])
            elif line.command in ("PART", "KICK"):
                log.debug("Removing %s from %s members", sent.left[0].username, sent.channel.source)
                self._members[sent.channel.source].remove(sent.left[0])
        elif line.command == "QUIT":
            nick = line.source.split("!", 1)[0]
            find = immp.User(id_=line.source, plug=self, username=nick)
            for name, members in list(self._members.items()):
                if name == nick:
                    log.debug("Removing %s self entry", nick)
                    del self._members[nick]
                elif find in members:
                    log.debug("Converting QUIT to PART for %s in %s", nick, name)
                    await self._handle(Line("PART", name, source=line.source))
        elif line.command == "NICK":
            old, host = line.source.split("!", 1)
            new = line.args[0]
            find = immp.User(id_=line.source, plug=self, username=old)
            replace = immp.User(id_="{}!{}".format(new, host), plug=self, username=new, raw=line)
            for name, members in list(self._members.items()):
                if name == old:
                    log.debug("Replacing %s with %s in self entry", old, new)
                    del self._members[old]
                    self._members[new] = {replace}
                elif find in members:
                    log.debug("Replacing %s with %s in %s members", old, new, name)
                    members.remove(find)
                    members.add(replace)
        elif line.command == "INVITE" and self.config["accept-invites"]:
            await self._join(line.args[1])
        elif line.command == "005":
            for param in line.args[1:]:
                if param.startswith("PREFIX="):
                    self._prefixes = param.split(")", 1)[1]
        elif line.command == "433":
            # Nickname in use, try another.
            self.config["user"]["nick"] += "_"
            self.write(Line("NICK", self.config["user"]["nick"]),
                       Line("USER", "immp", "0", "*", self.config["user"]["real-name"]))

    def wait(self, *lines, success=(), fail=(), collect=()):
        wait = Wait(lines, success, fail, collect)
        log.debug("Adding wait: %r", wait)
        self._waits.put_nowait(wait)
        self._sync_wait()
        return wait

    def write(self, *lines):
        for line in lines:
            log.debug("Sending line: %r", line)
            self._writer.write("{}\r\n".format(line).encode())

    def _lines(self, rich, user, action, edited):
        if not rich:
            return []
        elif not isinstance(rich, immp.RichText):
            rich = immp.RichText([immp.Segment(rich)])
        lines = []
        for text in IRCRichText.to_formatted(rich).split("\n"):
            if user:
                template = "* {} {}" if action else "<{}> {}"
                text = template.format(user.username or user.real_name, text)
            if edited:
                text = "[edit] {}".format(text)
            if not user and action:
                text = "\x01ACTION {}\x01".format(text)
            lines.append(text)
        return lines

    async def put(self, channel, msg):
        lines = []
        edited = msg.edited if isinstance(msg, immp.Receipt) else False
        if msg.text:
            lines += self._lines(msg.text, msg.user, msg.action, edited)
        for attach in msg.attachments:
            if isinstance(attach, immp.File):
                text = "uploaded a file{}".format(": {}".format(attach) if str(attach) else "")
                lines += self._lines(text, msg.user, True, edited)
            elif isinstance(attach, immp.Location):
                text = "shared a location: {}".format(attach)
                lines += self._lines(text, msg.user, True, edited)
            elif isinstance(attach, immp.Message) and attach.text:
                lines += self._lines(attach.text, attach.user, attach.action,
                                     attach.edited if isinstance(attach, immp.Receipt) else False)
        ids = []
        for text in lines:
            line = Line("PRIVMSG", channel.source, text)
            self.write(line)
            line.source = self._source
            sent = await IRCMessage.from_line(self, line)
            self.queue(sent)
            ids.append(sent.id)
        return ids



from .schema import Optional, Schema
from .util import ConfigProperty, Configurable, pretty_str


_GROUP_FIELDS = ("channels", "anywhere", "named", "private", "shared")


class _Schema:

    group = Schema({Optional(field, list): [str] for field in _GROUP_FIELDS})


@pretty_str
class Channel:
    """
    Container class that holds a (:class:`.Plug`, :class:`str`) pair representing a room
    inside the plug's network.

    Attributes:
        plug (.Plug):
            Related plug instance where the channel resides.
        source (str):
            Plug-specific channel identifier.
    """

    __slots__ = ("plug", "source")

    def __init__(self, plug, source):
        self.plug = plug
        self.source = str(source)

    async def is_private(self):
        """
        Equivalent to :meth:`.Plug.channel_is_private`.

        Returns:
            bool:
                ``True`` if the channel is private; ``None`` if the service doesn't have a notion
                of private channels.
        """
        return await self.plug.channel_is_private(self)

    async def title(self):
        """
        Equivalent to :meth:`.Plug.channel_title`.

        Returns:
            str:
                Display name for the channel.
        """
        return await self.plug.channel_title(self)

    async def link(self):
        """
        Equivalent to :meth:`.Plug.channel_link`.

        Returns:
            str:
                Internal deep link to this channel.
        """
        return await self.plug.channel_link(self)

    async def rename(self, title):
        """
        Equivalent to :meth:`.Plug.channel_rename`.

        Args:
            title (str):
                New display name for the channel.
        """
        return await self.plug.channel_rename(self, title)

    async def members(self):
        """
        Equivalent to :meth:`.Plug.channel_members`.

        Returns:
            .User list:
                Members present in the channel.
        """
        return await self.plug.channel_members(self)

    async def invite(self, user):
        """
        Equivalent to :meth:`.Plug.channel_invite`.

        Args:
            user (.User):
                New user to invite.
        """
        return await self.plug.channel_invite(self, user)

    async def remove(self, user):
        """
        Equivalent to :meth:`.Plug.channel_remove`.

        Args:
            user (.User):
                Existing user to kick.
        """
        return await self.plug.channel_remove(self, user)

    async def history(self, before=None):
        """
        Equivalent to :meth:`.Plug.channel_history`.

        Args:
            before (.Receipt):
                Starting point message, or ``None`` to fetch the most recent.

        Returns:
            .Receipt list:
                Messages from the channel, oldest first.
        """
        return await self.plug.channel_history(self, before)

    async def send(self, msg):
        """
        Push a message object to the related plug on this channel.

        Args:
            msg (.Message):
                Original message received from another channel or plug.
        """
        return await self.plug.send(self, msg)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.plug == other.plug and self.source == other.source)

    def __hash__(self):
        return hash((self.plug.network_id, self.source))

    def __repr__(self):
        return "<{}: {} @ {}>".format(self.__class__.__name__, self.plug.name, self.source)


@pretty_str
class Group(Configurable):
    """
    Container of multiple channels.

    Groups cannot be iterated, as they may hold any possible channel from a :class:`.Plug`, but you
    can test for membership of a given :class:`.Channel` using :meth:`has`.

    A group is defined by a base list of channels, and/or lists of channel types from plugs.  The
    latter may target **private** or **shared** (non-private) channels, **named** for host-defined
    channels, or **anywhere** as long as it belongs to the given plug.
    """

    class MergedProperty(ConfigProperty):

        def __init__(self, key=None):
            super().__init__([Group], key)

        def __get__(self, instance, owner):
            return Group.merge(instance.host, *super().__get__(instance, owner))

    schema = _Schema.group

    _channels = ConfigProperty([Channel])

    @classmethod
    def merge(cls, host, *groups):
        config = {field: [] for field in _GROUP_FIELDS}
        for group in groups:
            for field in _GROUP_FIELDS:
                config[field].extend(item for item in group.config[field]
                                     if item not in config[field])
        return cls(None, config, host)

    async def has_channel(self, channel):
        if not isinstance(channel, Channel):
            raise TypeError
        elif self._channels and channel in self._channels:
            return True
        elif self.has_plug(channel.plug, "anywhere"):
            return True
        elif self.has_plug(channel.plug, "named") and channel in self.host.channels.values():
            return True
        private = await channel.is_private()
        if self.has_plug(channel.plug, "private") and private:
            return True
        elif self.has_plug(channel.plug, "shared") and not private:
            return True
        else:
            return False

    def has_plug(self, plug, *fields):
        return any(plug.name in self.config[field] for field in fields or _GROUP_FIELDS)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)
